#include "slicing.h"
#include "stdlib.h"
#include "stdio.h"
#include "assert.h"

#define MAX_INSTR_GID 1000000
#define II_TABLE_SIZE 10000
instr_info * iis[MAX_INSTR_GID];
static instr_info * iis_table[II_TABLE_SIZE];
Int instr_gid=0;
static traced_file *traced_file_table[N_FILE_ENTRIES];
#define MAX_ERROR 10
error_node err_nodes[MAX_ERROR];
Int err_num;
Int err_found[MAX_ERROR];
Int always_traced=0;
Int predicate_size;
#define MAX_SOURCE_CODE_LINE 150000
Char is_bracket[MAX_SOURCE_CODE_LINE];
Int source_line_cnt;
void load_error_file ( Char* filename )
{
#  define N_BUF 200
   Int   fd;
   Char  buf[N_BUF+1];
   Int i=0,j;  
   fd = fopen( filename, "r" );
   if (fd < 0) {
      printf("FATAL: can't open error file `%s'",filename );
      exit(-1);
   }

   while (fgets(buf, N_BUF, fd)) {
      j=0;
      while(buf[j]!=' ') j++;
      buf[j]='\0';
      strcpy(err_nodes[i].file, buf);
      err_nodes[i].line=atoi (buf+j+1);
      printf("XXXX,%s,%d\n",err_nodes[i].file, err_nodes[i].line);
      i++;
   }
   err_num=i;
}

static Bool is_worklist_contain_error(slice_node * wl)
{
   slice_node * temp=wl;
   Int i;
   while (temp) {
      for (i=0;i<err_num;i++) {
	 if (err_found[i]) continue;
	 if (!strcmp(temp->ii->file, err_nodes[i].file) && temp->ii->line==err_nodes[i].line) {
             printf("XXXX contain error\n");
             err_found[i]=1;
         }
      }
      temp=temp->next;
   }
   for (i=0;i<err_num;i++) {
      if (!err_found[i]) return False;
   }
   return True;
}

static UInt hash(Char *s, UInt table_size)
{
   const int hash_constant = 256;
   int hash_value = 0;
   for ( ; *s; s++)
      hash_value = (hash_constant * hash_value + *s) % table_size;
   return hash_value;
}


static __inline__
traced_file* new_traced_file(Char filename[], traced_file * next)
{
   // Using calloc() zeroes the fns[] array
   traced_file* tf = calloc(1, sizeof(traced_file));
   tf->file   = strdup(filename);
   if (!strcmp(filename,"rfc822.c")) printf("\t%s --> %s\n",filename, tf->file);
   tf->next   = next;
   return tf;
}

static Bool is_file_traced(Char * file, Char * fn, Int line)
{
#define CONSIDER_INPUT
#ifdef CONSIDER_INPUT
      if (always_traced) return True;
      //if (!strcmp(fn,"_IO_getline_info")) return True;
      if (!strcmp(fn,"memcpy")) return True; //for buffer
      if (!strcmp(fn,"strcpy")) return True; //for buffer
      if (!strcmp(fn,"yyparse")) return True; //for bc
      if (!strcmp(fn,"yylex")) return True; //for bc
      if (!strcmp(fn,"input")) return True; //for bc
      if (!strcmp(fn,"getenv")) return True; //for tar
#endif
      traced_file * curr_traced_file;
      Int file_hash;
      file_hash = hash(file, N_FILE_ENTRIES);
      curr_traced_file   = traced_file_table[file_hash];
      
      while (NULL != curr_traced_file && strcmp(file, curr_traced_file->file)) {
         curr_traced_file = curr_traced_file->next;
      }
      
      return (curr_traced_file!=NULL&& !curr_traced_file->is_bracket[line]);
}

void load_traced_file ( Char* filename )
{
#  define N_BUF 4096 
   Int   fd,t_fd;
   Bool  eof;
   Char  buf[N_BUF+1];
   UInt file_hash;
   traced_file * curr_traced_file;
   Char c,only_bracket;
   Int i;
   fd = fopen( filename, "r" );
   if (fd < 0) {
      printf("FATAL: can't open traced file `%s'",filename );
      exit(-1);
   }

   while (fgets(buf, N_BUF, fd)) {
      buf[strlen(buf)-1]='\0';
      printf("LOADED %s\n", buf);
      /* Assign and initialise the two suppression halves (core and tool) */
      file_hash = hash(buf, N_FILE_ENTRIES);
      printf("\t%s\n",buf);
      //if (!strcmp(buf,"rfc822.c")) printf("\tHASH=%d\n",file_hash);      
      curr_traced_file   = traced_file_table[file_hash];
      //if (curr_traced_file && !strcmp(buf,"rfc822.c")) printf("\ttest %s\n",curr_traced_file->file);      
      while (NULL != curr_traced_file && strcmp(buf, curr_traced_file->file)) {
         //if (curr_traced_file && !strcmp(buf,"rfc822.c")) printf("\ttest %s\n",curr_traced_file->file);      
         curr_traced_file = curr_traced_file->next;
      }
      if (NULL == curr_traced_file) {
         //if (!strcmp(buf,"rfc822.c")) printf("\tDUPLICATE\n");      
         curr_traced_file =
             new_traced_file(buf, traced_file_table[file_hash]);
         //if (!strcmp(buf,"rfc822.c")) printf("\tDUPLICATE RESULT %s\n",curr_traced_file->file);      
         traced_file_table[file_hash] = curr_traced_file;
         //distinct_files++;
      }
      source_line_cnt=1;
      t_fd=fopen(buf,"r");
      if (t_fd) {
	  int j=0;
          while (fgets(buf, N_BUF,t_fd)) {
	     only_bracket=1;
             for(i=0;i<strlen(buf);i++) {
		c=buf[i];
		if (c!=' ' && c!='\t' && c!='\n') {
		   if (c!='{' && c!='}') {
		      only_bracket=0;
		      break;
		   }  
		}
	     }//for  
             is_bracket[source_line_cnt++]=only_bracket;
	     //printf("%d(%d):%s",source_line_cnt,only_bracket,buf);
          }
	  fclose(t_fd);
	  curr_traced_file->is_bracket=malloc(source_line_cnt);
          memcpy(curr_traced_file->is_bracket, is_bracket, source_line_cnt);
      }//if
   }//while
}


static val_stream * create_val_stream_with_size(Int size)
{
    val_stream * vs;
    vs=calloc(1, sizeof(val_stream));
    vs->allocated=vs->used=size;
    vs->iis=calloc(vs->allocated, sizeof(UInt));
    vs->insts=calloc(vs->allocated, sizeof(UInt));
    //vs->visited=calloc(vs->allocated/32+1, sizeof(UInt));
    return vs;
}

static void set_forward_visited_bit(instr_info * ii, Int inst)
{
    //printf("SET %x, %d \n", vs, inst);
    Int word_idx=inst/32;
    Int bit_idx=inst%32;
    UInt bit_mask=1;
    bit_mask=bit_mask<<bit_idx;
    ii->forward_visited[word_idx]=ii->forward_visited[word_idx] | bit_mask;
    //printf("\t SET %x,%x\n",vs->visited[word_idx],bit_mask);
}

Int get_forward_visited_bit(instr_info * ii, Int inst)
{
    //printf("GET ii %d, inst= %d cur_inst=%d\n",ii->id,inst,ii->cur_inst);
    if (!ii->forward_inst || inst>ii->forward_inst) {
	//printf("NULL visited, with alloced=%d\n",vs->allocated);
	return True;
    }
    Int word_idx=inst/32;
    UInt bit_mask= 1;
    bit_mask = (bit_mask<< (inst %32));
    //printf("\t GET %x, %x\n",vs->visited[word_idx],bit_mask);
    return ii->forward_visited[word_idx] & bit_mask;
}

static void reset_vs_visited_bit(instr_info * ii, Int inst)
{
    Int word_idx=inst/32;
    Int bit_idx=inst%32;
    UInt bit_mask=1;
    bit_mask=bit_mask<<bit_idx;
    ii->visited[word_idx]=ii->visited[word_idx] & (~bit_mask);
}

static void set_vs_visited_bit(instr_info * ii, Int inst)
{
    //printf("SET %d, %d \n", ii->id, inst);
    Int word_idx=inst/32;
    Int bit_idx=inst%32;
    UInt bit_mask=1;
    bit_mask=bit_mask<<bit_idx;
    ii->visited[word_idx]=ii->visited[word_idx] | bit_mask;
    //printf("\t SET %x,%x\n",ii->visited[word_idx],bit_mask);
}

static Int get_certified_bit(instr_info * ii, Int inst)
{
    if (!ii->cur_inst || inst>ii->cur_inst) {
	//printf("NULL visited, with alloced=%d\n",vs->allocated);
	return True;
    }
    Int word_idx=inst/32;
    UInt bit_mask= 1;
    bit_mask = (bit_mask<< (inst %32));
    //printf("\t GET %x, %x\n",ii->certified[word_idx],bit_mask);
    return ii->certified[word_idx] & bit_mask;
}

static void set_certified_bit(instr_info * ii, Int inst)
{
    //printf("SET %d, %d \n", ii->id, inst);
    Int word_idx=inst/32;
    Int bit_idx=inst%32;
    UInt bit_mask=1;
    bit_mask=bit_mask<<bit_idx;
    ii->certified[word_idx]=ii->certified[word_idx] | bit_mask;
    //printf("\t SET %x,%x\n",ii->certified[word_idx],bit_mask);
}

Int get_vs_visited_bit(instr_info * ii, Int inst)
{
    if (!ii->cur_inst || inst>ii->cur_inst) {
	//printf("NULL visited, with alloced=%d\n",vs->allocated);
	return True;
    }
    Int word_idx=inst/32;
    UInt bit_mask= 1;
    bit_mask = (bit_mask<< (inst %32));
    //printf("\t GET %x, %x\n",ii->visited[word_idx],bit_mask);
    return ii->visited[word_idx] & bit_mask;
}



#define CLEAR_VS_BITS(ii)  memset(ii->visited,0,(ii->cur_inst/32+1)*4)
#define CLEAR_CF_BITS(ii)  memset(ii->certified,0,(ii->cur_inst/32+1)*4)
void clear_all_certified_bits()
{
    Int i,j;
    instr_info * ii;
    val_stream * vs;

    for (i=0;i<instr_gid;i++) {
        ii=iis[i];
        if (ii->cur_inst) CLEAR_CF_BITS(ii);
    }
}
void clear_all_visited_bits()
{
    Int i,j;
    instr_info * ii;
    val_stream * vs;

    for (i=0;i<instr_gid;i++) {
        ii=iis[i];
        if (ii->cur_inst) CLEAR_VS_BITS(ii);
    }
}

#define CLEAR_FORWARD_VS_BITS(ii)  memset(ii->forward_visited,0,(ii->forward_inst/32+1)*4)

void clear_all_forward_visited_bits()
{
    Int i,j;
    instr_info * ii;
    val_stream * vs;

    for (i=0;i<instr_gid;i++) {
        ii=iis[i];
        if (ii->forward_inst) CLEAR_FORWARD_VS_BITS(ii);
    }
}

static instr_info * create_instr_info (Addr pc, Int bb_id, Char * fn, Char * file, Int line, Int id, Int uport, Addr jmp_pc)
{
     instr_info * ii;
     ii=calloc (1, sizeof(instr_info));

     ii->orig_addr=pc;
     ii->bb_id=bb_id;

     ii->fn=calloc (strlen(fn)+1, sizeof (Char));
     strcpy (ii->fn, fn);

     ii->file=calloc (strlen(file)+1, sizeof (Char));
     strcpy (ii->file, file);

     ii->line=line;
     ii->id=id;
     ii->use_port=uport;
     ii->jmp_pc=jmp_pc;
     return ii;
}

void print_instr(instr_info * ii)
{
    printf("%d: (%x, %s, %s, %d) ",ii->id, ii->orig_addr, ii->file, ii->fn, ii->line);
}


slice_node *  get_criteria_from_output (Char * sim_output, Int pid, Int err_id, Int err_inst)
{
    FILE * fd;
    Int id, inst,i,offset;
    Char buf[256];
    Char fn[128];
    Char file[128];
    Addr pc;
    instr_info * ii;
    slice_node * criteria=NULL;
    fd = fopen( sim_output, "r" );
    if (fd < 0) {
       printf("FATAL: can't open simplified output file `%s'", sim_output);
       exit(-1);
    }
    Int process_found=0;
    while (fgets(buf,256,fd)) {
	printf("%s",buf);
	if (buf[1]=='P' && buf[2]=='I' && buf[3]=='D') {
            //if (atoi(buf+5)==pid) process_found=1;
            if (atoi(buf+5)!=pid) process_found=1;
	    else process_found=0;
        } else if (process_found) {
	   if (buf[0]='&' && buf[1]=='*') {
	       i=3;
	       while (buf[i]!=',') i++;
	       buf[i]='\0';
	       id=atoi(buf+3);
	       inst=atoi(buf+i+1);
               if (err_id!=id || inst !=err_inst) {
	 	   printf("insert <%d,%d>\n",id,inst);
	   	   criteria=insert_node_into_list(iis[id],inst,criteria);
	       } else {
		   fclose(fd);
		   return criteria;
	       }
           }
	}
     }//while
     fclose(fd);
     return criteria;
	    
}

slice_node *  get_dice_criteria_with_index (Char * sim_output,Int index)
//slice_node *  get_dice_criteria (Char * sim_output, Int err_id, Int err_inst)
{
    FILE * fd;
    Int id, inst,i,offset;
    Char buf[1024];
    Char fn[128];
    Char file[128];
    Addr pc;
    instr_info * ii;
    slice_node * criteria=NULL;
    Int idx;
    fd = fopen( sim_output, "r" );
    if (fd < 0) {
       printf("FATAL: can't open simplified output file `%s'", sim_output);
       exit(-1);
    }
    idx=0;
    while (fgets(buf,1024,fd) && idx<index) {
	if (buf[0]=='&' && buf[1]=='*') {
	    i=3;
	    while (buf[i]!=',') i++;
	    buf[i]='\0';
	    id=atoi(buf+3);
	    inst=atoi(buf+i+1);
            //if (err_id!=id || inst !=err_inst) {
		criteria=insert_node_into_list(iis[id],inst,criteria);
	    //} else {
		//fclose(fd);
		//return criteria;
	    //}
        }
	idx++;
     }//while
     fclose(fd);
     return criteria;
}

//slice_node *  get_dice_criteria (Char * sim_output)
slice_node *  get_dice_criteria_with_err (Char * sim_output, Int err_id, Int err_inst)
{
    FILE * fd;
    Int id, inst,i,offset;
    Char buf[1024];
    Char fn[128];
    Char file[128];
    Addr pc;
    instr_info * ii;
    slice_node * criteria=NULL;
    printf("%s,%d,%d\n",sim_output,err_id,err_inst);
    fd = fopen( sim_output, "r" );
    if (fd < 0) {
       printf("FATAL: can't open simplified output file `%s'", sim_output);
       exit(-1);
    }
    
    while (fgets(buf,1024,fd)) {
	if (buf[0]=='&' && buf[1]=='*') {
	    i=3;
	    while (buf[i]!=',') i++;
	    buf[i]='\0';
	    id=atoi(buf+3);
	    inst=atoi(buf+i+1);
            if (err_id==id && inst ==err_inst) {
		fclose(fd);
		return criteria;
	    }
  	    criteria=insert_node_into_list(iis[id],inst,criteria);
        }
     }//while
     fclose(fd);
     return criteria;
}

slice_node *  get_dice_criteria (Char * sim_output)
//slice_node *  get_dice_criteria (Char * sim_output, Int err_id, Int err_inst)
{
    FILE * fd;
    Int id, inst,i,offset;
    Char buf[1024];
    Char fn[128];
    Char file[128];
    Addr pc;
    instr_info * ii;
    slice_node * criteria=NULL;
    fd = fopen( sim_output, "r" );
    if (fd < 0) {
       printf("FATAL: can't open simplified output file `%s'", sim_output);
       exit(-1);
    }
    
    while (fgets(buf,1024,fd)) {
	if (buf[0]=='&' && buf[1]=='*') {
	    i=3;
	    while (buf[i]!=',') i++;
	    buf[i]='\0';
	    id=atoi(buf+3);
	    inst=atoi(buf+i+1);
            //if (err_id!=id || inst !=err_inst) {
		criteria=insert_node_into_list(iis[id],inst,criteria);
	    //} else {
		//fclose(fd);
		//return criteria;
	    //}
        }
     }//while
     fclose(fd);
     return criteria;
}

slice_node * find_input_criteria(Char * inputs)
{
    Int len=strlen(inputs);
    Int i,j,k;
    slice_node * result=NULL;
    instr_info * ii;
    for (i=0;i<instr_gid;i++) {
	ii=iis[i];
	if (ii->input_related) {
           for (j=0;j<ii->cur_inst;j++) {
	     for (k=0;k<len;k++) {
                if (ii->input_related[j]==inputs[k]) {
		   result=insert_node_into_list(ii,j,result);
		   break;
		}
	     }//k
	   }//j
	}//if
    }
    return result;
} 

  	   
void read_instr_dic (Char * dic)
{
    FILE * fd;
    Int i, offset,eof,line,id,uport,bb_id,pred_id;
    Char buf[256];
    Char fn[128];
    Char file[128];
    Addr pc,jmp_pc;
    instr_info * ii;
    Char c;
    Bool is_predicate;
    fd = fopen( dic, "r" );
    if (fd < 0) {
       printf("FATAL: can't open traced file `%s'", dic);
       exit(-1);
    }

    fgets (buf,256,fd);
    instr_gid=atoll (buf);
    //instr_gid=0;
    fgets (buf,256,fd);
    jmp_pc=0;
    while (fgets(buf,256,fd)) {
      pred_id=-1;
      /* Assign and initialise the two suppression halves (core and tool) */
      i=0;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      id=atoll(buf+offset);

      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      pc=strtoll(buf+offset,NULL,16);


      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      bb_id=atoll(buf+offset);

      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      strcpy(fn, buf+offset);

      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      strcpy(file, buf+offset);

      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      line=atoll(buf+offset);

      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      //for backward compatibility after JMP_PC is added
      //if (buf[i]==' ') jmp_pc=strtoll(buf+i+1,NULL,16);
      c=buf[i];
      buf[i]='\0';
      uport=atoll(buf+offset);

      if (c!='\n') {
	i++;
        offset=i;
	while (buf[i]!=' ' && buf[i]!='\n') i++;
	c=buf[i];
	buf[i]='\0';
        jmp_pc=strtoll(buf+offset,NULL,16);
	if (c!='\n') {
	   //is_predicate=atoll(buf+i+1);
           i++;
           offset=i;
	   while (buf[i]!=' ' && buf[i]!='\n') i++;
	   c=buf[i];
	   buf[i]='\0';
           is_predicate=atoll(buf+offset);
	   if (c!='\n') {
	      pred_id=atoll(buf+i+1);
           }
        }
      }


      iis[id]=create_instr_info(pc, bb_id, fn, file, line, id, uport, jmp_pc);
      iis[id]->is_predicate=is_predicate;
      iis[id]->pred_id=pred_id;
      //put it into hash table
      Int index=pc%II_TABLE_SIZE;
      iis[id]->next=iis_table[index];
      iis_table[index]=iis[id];
      //print_instr(iis[id]);      
   }
   fclose(fd);

   for (i=0;i<instr_gid;i++) {
       ii=iis[i];
       if (ii->bb_id!=-1) ii->bb=iis[ii->bb_id];
 
       if (ii->pred_id!=-1) {
	   ii->pred_ii=iis[ii->pred_id];
       }
   }
}

static val_stream * read_val_stream (fd)
{
   val_stream * vs;
   Int size,i,j,offset, ii_id;
   Char buf[256];

   fgets (buf,256,fd);//"SIZE %d"
   //printf("READ_VAL%s",buf);
   size=atoll (buf+5);
   if (!size) return NULL;
   vs=create_val_stream_with_size(size);

   fgets (buf,256,fd);//"II INST"

   for (i=0;i<size;i++) {
      fgets (buf,256,fd);//"%d %d"

      j=1;
      while (buf[j]!=':') j++;
      offset=j+1;
      while (buf[j]!=' ' && buf[j]!='\n') j++;
      buf[j]='\0';
      ii_id=atoll (buf+offset);
      if (ii_id==-1) vs->iis[i]=NULL;
      else vs->iis[i]=(UInt)iis[ii_id];
      vs->insts[i]=atoll (buf+j+1);
   }
   return vs;
}

#define MAX_PCD 512
instr_info * get_instr_info_by_pc(Addr pc)
{
    instr_info * ii;
    if (!pc) return NULL;
    Int index=pc%II_TABLE_SIZE;
    ii=iis_table[index];
    while (ii && ii->orig_addr !=pc) ii=ii->next;
    return ii;
}

void read_potential_cd(Char * file)
{
    FILE * fd;
    Char buf[8*MAX_PCD];
    Addr load_pc, pred_pc, branch_pc;
    Int i,j, offset,idx;
    instr_info * pred_bbs[MAX_PCD];
    instr_info * branches[MAX_PCD];
    instr_info * ii_load, * ii_pred, * ii_branch;
    potential_cd * pcd;
    

    fd = fopen( file, "r" );
    if (fd < 0) {
       printf("FATAL: can't open traced file `%s'", file);
       exit(-1);
    }

    while (fgets(buf,4096,fd)) {
      i=0;
      offset=i;
      while (buf[i]!=':' && buf[i]!='\0') i++;
      assert(buf[i]!='\0');
      buf[i]='\0';
      load_pc = strtol(buf+offset,NULL,16);
      //printf("load_pc=%x\n", load_pc);
      ii_load=get_instr_info_by_pc(load_pc);
      //this load is not executed in the considered run
      if (!ii_load) continue;
      //printf("for ii_load %x (%s, %d)\n", ii_load->orig_addr, ii_load->file, ii_load->file);
      idx=0;
      i+=2;
      offset=i;
      
      //parse the predicates and branches
      while (buf[i]!='\n' && buf[i]!='\0') {
        while (buf[i]!='-' && buf[i]!='\0' && buf[i]!='\n') i++;
	assert(buf[i]=='-');
	buf[i]='\0';
        pred_pc = strtol(buf+offset, NULL, 16);
        ii_pred=get_instr_info_by_pc(pred_pc);

        i++;
	offset=i;
        while (buf[i]!=' ' && buf[i]!='\n' && buf[i]!='\0') i++;
	assert(buf[i]==' ');
        buf[i]='\0';
        branch_pc=strtol(buf+offset, NULL,16);
	ii_branch=get_instr_info_by_pc(branch_pc);

        if (ii_pred) {
	   pred_bbs[idx]=ii_pred;
	   branches[idx++]=ii_branch;
	}
	i++;
	offset=i;
     }//while
     if (idx) {
	printf("ii %x(%s,%d) has %d PCD\n", ii_load->orig_addr, ii_load->file, ii_load->line, idx);
	ii_load->pcd=pcd= calloc(1, sizeof(potential_cd));
	pcd->load_ii=ii_load;
        pcd->pcd_cnt=idx;
        pcd->p_pred_bbs=calloc(idx, sizeof (instr_info*));
        pcd->p_branches=calloc(idx, sizeof (instr_info*));
        for (j=0;j<idx;j++) {
	   pcd->p_pred_bbs[j]=pred_bbs[j];
	   pcd->p_branches[j]=branches[j];
	}
     }
   }//while
   fclose(fd);
}

//for relevant slicing, for outputting the potential dependence
void write_potential_dependence(Char * file)
{
   FILE * fd;
   Int i,j;
   instr_info * ii;
   potential_dep * pd;
   fd = fopen( file, "w" );
   if (fd < 0) {
       printf("FATAL: can't open traced file `%s'", file);
       exit(-1);
   }
   
   fprintf(fd,"%d\n",instr_gid);
   for (i=0;i<instr_gid;i++) {
      ii=iis[i];
      fprintf(fd,"%d %x %s %d\n", ii->id,ii->orig_addr,ii->file, ii->line);
      fprintf(fd,"PDS %d\n",ii->pds_alloced);
      if (ii->pds_alloced) {
         for (j=0;j<ii->pds_alloced;j++) {
	    if (ii->pds[j]) {
		fprintf(fd,"\t%d: YES\n",j);
		pd=ii->pds[j];
                while (pd) {
		   if (pd->pred_ii) 
                       fprintf(fd, "\t\t%d %d\n",pd->pred_ii->id, pd->pred_inst);
                   pd=pd->next;
                }
	        fprintf(fd,"\t\tEND\n");
	     } else {
		fprintf(fd,"\t%d: NO\n",j);
	     }
	 }//j
      }//if
      //write_val_stream(fd,ii->pds);
   }
}
//for relevant slicing
void read_potential_dependence(Char * file)
{
   FILE * fd;
   Int i,j,id,pds_cnt,offset,pred_id, pred_inst;
   instr_info * ii;
   Char buf[256];
   potential_dep * pd;
   
   fd = fopen( file, "r" );
   if (fd < 0) {
       printf("FATAL: can't open traced file `%s'", file);
       exit(-1);
   }
   
   //fprintf(fd,"%d\n",instr_gid);
   fgets(buf,256,fd);
   //for (i=0;i<instr_gid;i++) {
   while (fgets(buf,256,fd)) {
      id=atoi(buf);
      
      ii=iis[id];
      fgets(buf,256,fd);
      sscanf(buf, "PDS %d\n", &pds_cnt);
      printf("id=%d, pds %d\n",id, pds_cnt);
      ii->pds_alloced=pds_cnt;
      if (pds_cnt) {
	 ii->pds=calloc(pds_cnt, sizeof(potential_dep*));
         for (j=0;j<pds_cnt;j++) {
	    fgets(buf,256,fd);
	    offset=1;
	    while (buf[offset]!=' ') offset++;
	    if (buf[offset+1]=='Y') {
                fgets(buf,256,fd);
	        while (buf[2]!='E') {
		   pd=calloc(1, sizeof (potential_dep));
		   sscanf(buf,"\t\t%d %d\n", &pred_id, & pred_inst);
		   pd->pred_ii=iis[pred_id];
		   pd->pred_inst=pred_inst;
                   pd->next=ii->pds[j];		
		   ii->pds[j]=pd;
		   fgets(buf,256,fd);
               }
		//ii->pds[j]=read_pd_one_entry(fd);
	    } //else for 'N'
         }//for j
      }//if
   }//for i
   fclose(fd);
}
Int total_size=0;
Int instr_total=0;
Int traced_instr_total=0;
void read_dependence (Char * file)
{
    FILE * fd;
    Int i, offset,eof,id,uport,j;
    Char buf[256];
    instr_info * ii,*cd_ii;
    Int cur_inst,cnt;
    val_stream * cd_vs;
    fd = fopen( file, "r" );
    if (fd < 0) {
       printf("FATAL: can't open traced file `%s'", file);
       exit(-1);
    }
    fgets(buf,256,fd); //instr_gid
    //instr_gid=VG_(atoll) (buf);

    fgets(buf,256,fd); //"ID USE_PORT"

    while (fgets(buf,256,fd)) {
      /* Assign and initialise the two suppression halves (core and tool) */
      i=0;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      id=atoll(buf+offset);

      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      uport=atoll(buf+offset);

      ii=iis[id];
      ii->cur_inst=0;
      for (j=0;j<uport;j++) {
          //printf("id=%d, uport=%d\n", id, uport);
          ii->uports[j]=read_val_stream(fd);
          //if (j==0 && ii->uports[j]) printf("I got CD\n");
          if (ii->uports[j]) total_size+=ii->uports[j]->used;
	  if (ii->uports[j] && ii->uports[j]->used > ii->cur_inst) {
	     ii->cur_inst=ii->uports[j]->used;
          }
      }
      fgets(buf,256,fd); //VALUES or NO VALUES or "INPUT and VALUES"
      cnt=0;
      switch (buf[0]) {
      case 'V':
         cnt=atoi(buf+7);
         ii->values=calloc(cnt, sizeof(UInt));
         ii->allocated=cnt;
         for (j=0;j<cnt;j++) {
            fgets(buf,256,fd);
            i=1;
            while (buf[i]!=':') i++;
            ii->values[j]=strtol(buf+i+1,NULL,16);
	 }
	 break;
       case 'I':
         cnt=atoi(buf+7);
         ii->input_related=calloc(cnt, sizeof(UInt));
         for (j=0;j<cnt;j++) {
            fgets(buf,256,fd);
	    //printf("values for %d =%s",id,buf);
            i=1;
            while (buf[i]!=':') i++;
            ii->input_related[j]=atoi(buf+i+1);
	 }
	 break;
       case 'B':
	 cnt=atoi(buf+5);

         ii->allocated=cnt;
         ii->values=calloc(cnt, sizeof(UInt));
         ii->input_related=calloc(cnt,sizeof(UInt));
         for (j=0;j<cnt;j++) {
            fgets(buf,256,fd);
	    //printf("values for %d =%s",id,buf);
            i=1;
            while (buf[i]!=':') i++;
	    ii->input_related[j]=atoi(buf+i+1);
	    while (buf[i]!=' ') i++;
	    ii->values[j]=strtol(buf+i+1,NULL,16);
         }//for
      }//switch


      fgets(buf,256,fd); //TSS or NO TSS 
      cnt=0;
      if (buf[0]=='T') {
         cnt=atoi(buf+4);
         ii->tss=calloc(cnt, sizeof(UInt));
         ii->cfs=calloc(cnt, sizeof(UInt));
         ii->ts_alloced=cnt;
         for (j=0;j<cnt;j++) {
            fgets(buf,256,fd);
	    //printf("values for %d =%s",id,buf);
            i=1;
            while (buf[i]!=':') i++;
            ii->tss[j]=strtol(buf+i+1,NULL,16);
	    while (buf[i]!=' ') i++;
	    ii->cfs[j]=strtol(buf+i+1,NULL,16);
	 }
      }

      //set is_predicate 
#ifndef MAPPING
      if (ii->uports) {
	  cd_vs=ii->uports[0];
	  if (cd_vs) {
	     for (j=0;j<cd_vs->used;j++) {
	        cd_ii=cd_vs->iis[j];
		if (cd_ii) cd_ii->is_predicate=True;
             }
          }
      }
#endif

      if (ii->cur_inst) {
	 ii->visited=calloc(ii->cur_inst/32+1, sizeof(UInt));
#ifdef DEP_DISTANCE
         ii->back_dist=calloc(ii->cur_inst,sizeof(UInt));
#endif

#ifdef CERTIFY
	 ii->certified=calloc(ii->cur_inst/32+1, sizeof(UInt));
#endif
      }
      instr_total+=ii->cur_inst;
      if (is_file_traced(ii->file,ii->fn,ii->line)) traced_instr_total+=ii->cur_inst;
    //vs->visited=calloc(vs->allocated/32+1, sizeof(UInt));
   }//while
   printf("TOTAL instruction is %d(%d traced) total_size=%d(%d)\n", instr_total,traced_instr_total,total_size, total_size*4);
}


static slice_node * create_slice_node (instr_info * ii, Int inst)
{
    slice_node * n;
    n=calloc(1, sizeof(slice_node));
    n->ii=ii;
    n->inst=inst;
    if (ii->values) {
	n->value=ii->values[inst];
    } else {
	n->value=0xfffe;
    }
    return n;
}

static slice_node * create_slice_node_with_value (instr_info * ii, Int inst, UInt v)
{
    slice_node * n;
    n=calloc(1, sizeof(slice_node));
    n->ii=ii;
    n->inst=inst;
    n->value=v;
    return n;
}

static slice_node * create_slice_node_with_value_step (instr_info * ii, Int inst, UInt v,Int step)
{
    slice_node * n;
    n=calloc(1, sizeof(slice_node));
    n->ii=ii;
    n->inst=inst;
    n->value=v;
    n->step=step;
    return n;
}

static slice_node *  insert_node_into_list_tail (instr_info * ii,Int inst, slice_node * work_list)
{
    slice_node * node =create_slice_node(ii,inst);
    if (work_list) work_list->next=node;
    return node;
}

static slice_node *  insert_node_into_list_tail_with_value (instr_info * ii,Int inst, UInt v,slice_node * work_list)
{
    slice_node * node =create_slice_node_with_value(ii,inst,v);
    if (work_list) work_list->next=node;
    return node;
}

static slice_node *  insert_node_into_list_tail_with_value_step (instr_info * ii,Int inst, UInt v,Int step, slice_node * work_list)
{
    slice_node * node =create_slice_node_with_value_step(ii,inst,v,step);
    if (work_list) work_list->next=node;
    return node;
}
slice_node *  insert_node_into_list (instr_info * ii,Int inst, slice_node * work_list)
{
    slice_node * node =create_slice_node(ii,inst);
    node->next=work_list;
    return node;
}
static slice_node *  insert_node_into_list_with_value_step (instr_info * ii,Int inst, UInt value, Int step,slice_node * work_list)
{
    slice_node * node =create_slice_node_with_value_step(ii,inst,value,step);
    node->next=work_list;
    return node;
}

static int slice_node_compare_src_dd(const void * n1, const void * n2)
{
    slice_node * sn1=*((slice_node**)n1);
    slice_node * sn2=*((slice_node**)n2);
    Char * file1=sn1->ii->file;
    Char * file2=sn2->ii->file;
    Int r;
    if (!file1) return -1;
    if (!file2) return 1;
    r=strcmp(file1,file2);
    if (r>0) return 1;
    else if (r<0) return -1;
    Int line1=sn1->ii->line;
    Int line2=sn2->ii->line;
    if (line1<line2) return -1;
    else if (line1>line2) return 1;
   /* 
    UInt addr1=sn1->ii->orig_addr;
    UInt addr2=sn2->ii->orig_addr;
    if (addr1< addr2) return -1;
    if (addr1> addr2) return 1;
*/
    if (sn1->step<sn2->step) return -1;
    else return 1;
}

static int slice_node_compare_dd(const void * n1, const void * n2)
{
    slice_node * sn1=*((slice_node**)n1);
    slice_node * sn2=*((slice_node**)n2);
    if (sn1->step<sn2->step) return -1;
    else return 1;
}

slice_node * qsort_list_by_dd_compact(slice_node * list)
{
    Int cnt=0,i;
    slice_node * temp;
    slice_node ** nodes;
    temp=list;
    if (!temp) return temp;
    while (temp) {
      cnt++;
      temp=temp->next;
    }
	
    if (cnt==0) return NULL;
    nodes=malloc(sizeof(slice_node*)*cnt);
    temp=list;
    for (i=0;i<cnt;i++) {
       nodes[i]=temp;
       temp=temp->next;
    }
    qsort(nodes, cnt, sizeof(slice_node*), slice_node_compare_src_dd);
    //find the first instance for each statement (the one with the smallest step)   
    slice_node * old=nodes[0];
    Int idx=1;
    for (i=1;i<cnt-1;i++) {
	temp=nodes[i];
	if (old->ii->line!=temp->ii->line || (old->ii->file !=NULL && strcmp(old->ii->file, temp->ii->file))) {
            old->next=nodes[i];
	    old=nodes[i];
	    idx++;
	}
    }
    printf("IDX=%d\n",idx);
    old->next=NULL;
    temp=nodes[0];
    cnt=0;
    while (temp) {
      nodes[cnt++]=temp;
      temp=temp->next;
    }

    //sort them purely by dd  
    qsort(nodes, cnt, sizeof(slice_node*), slice_node_compare_dd);
    for (i=0;i<cnt-1;i++) {
      nodes[i]->next=nodes[i+1];
    }
    nodes[cnt-1]->next=NULL;
    return nodes[0];
}

static int slice_node_compare(const void * n1, const void * n2)
{
    slice_node * sn1=*((slice_node**)n1);
    slice_node * sn2=*((slice_node**)n2);
    UInt addr1=sn1->ii->orig_addr;
    UInt addr2=sn2->ii->orig_addr;
    if (addr1< addr2) return -1;
    if (addr1> addr2) return 1;
    if (sn1->inst<sn2->inst) return -1;
    else return 1;
}

slice_node * qsort_list(slice_node * list)
{
    Int cnt=0,i;
    slice_node * temp;
    slice_node ** nodes;
    temp=list;
    if (!temp) return temp;
    while (temp) {
      cnt++;
      temp=temp->next;
    }
	
    nodes=malloc(sizeof(slice_node*)*cnt);
    temp=list;
    for (i=0;i<cnt;i++) {
       nodes[i]=temp;
       temp=temp->next;
    }
    qsort(nodes, cnt, sizeof(slice_node*), slice_node_compare);
    for (i=0;i<cnt-1;i++) {
      nodes[i]->next=nodes[i+1];
    }
    nodes[cnt-1]->next=NULL;
    return nodes[0];
}

static int slice_node_compare_step_ts(const void * n1, const void * n2)
{
    slice_node * sn1=*((slice_node**)n1);
    slice_node * sn2=*((slice_node**)n2);
    Int s1=sn1->step;
    Int s2=sn2->step;
    //step increasing order
    if (s1< s2) return -1;
    if (s1> s2) return 1;
    //timestamp decreasing order
    if (sn1->value>sn2->value) return -1;
    else return 1;
}


slice_node * qsort_list_by_step_timestamp(slice_node * list)
{
    Int cnt=0,i;
    slice_node * temp;
    slice_node ** nodes;
    temp=list;
    while (temp) {
      cnt++;
      temp=temp->next;
    }
    nodes=malloc(sizeof(slice_node*)*cnt);
    temp=list;
    for (i=0;i<cnt;i++) {
       nodes[i]=temp;
       temp=temp->next;
    }
    qsort(nodes, cnt, sizeof(slice_node*), slice_node_compare_step_ts);
    for (i=0;i<cnt-1;i++) {
      nodes[i]->next=nodes[i+1];
    }
    nodes[cnt-1]->next=NULL;
    return nodes[0];
}


slice_node * sort_list(slice_node * list)
{
    slice_node * result=NULL;
    slice_node * si,*max, *prev_max,*prev_temp,*temp;
    si=list;
    instr_info * ii;
    while (si) {
      prev_max=NULL;
      max=si;
      prev_temp=NULL;
      temp=si;
      while (temp) {
        if (max->ii->orig_addr<temp->ii->orig_addr || (max->ii->orig_addr==temp->ii->orig_addr && max->inst<temp->inst)) {
	   prev_max=prev_temp;
           max=temp;
        }
        prev_temp=temp;
        temp=temp->next;
     }
     result=insert_node_into_list(max->ii, max->inst, result);
     if (prev_max) {
	prev_max->next=max->next;
     } else {
        si=max->next;
     }
     ii=max->ii; 
     printf("%d: (%x, %s, %s, %d),%d\n",ii->id, ii->orig_addr, ii->file, ii->fn, ii->line,max->inst);
     free(max);
  }//while(si)
  return result;
}   



slice_node *  insert_node_into_list_sorted (instr_info * ii,Int inst, slice_node * work_list)
{
    slice_node * head, * prev, *node;
    head=work_list;
    if (head==NULL) printf("work_list is NULL\n");
    prev=NULL;
    //rank by (ii->id, inst)
    while (work_list && work_list->ii->orig_addr > ii->orig_addr) {
    //while (work_list && work_list->ii->id > ii->id ) {
           prev=work_list;
           work_list=work_list->next;
    }

    while (work_list && work_list->ii->orig_addr==ii->orig_addr && work_list->inst> inst) {
    //while (work_list && work_list->ii->id==ii->id && work_list->inst> inst) {
           prev=work_list;
           work_list=work_list->next;
    }
    //node has bee in the list
    if (work_list && ii->orig_addr==work_list->ii->orig_addr && inst==work_list->inst) return head;
    //if (work_list && ii->id==work_list->ii->id && inst==work_list->inst) return head;
    //printf("CREATE (%x, %s, %s, %d), %d\n", ii->orig_addr, ii->file, ii->fn, ii->line, inst);
    node=create_slice_node(ii,inst);
    
    node->next=work_list;
    if (prev) prev->next=node;
    if (head==work_list) head=node;
    return head;
}

void slice_one_step_depth_first(slice_node * current_node, slice_node ** new_wl)
{
    instr_info * ii,*tmp_ii;
    Int inst,i,tmp_inst;
    val_stream * vs;
    instr_info * start_ii=current_node->ii;
    slice_node * old;
    slice_node *temp_wl=NULL;
    temp_wl=insert_node_into_list(start_ii, current_node->inst, temp_wl);
    //slice_node * temp_wl=creat_slice_node(start_ii, current_node->inst);
    while (temp_wl) {
       ii=temp_wl->ii;
       inst=temp_wl->inst;
       printf("\tprocess %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,temp_wl->value,0x7f&temp_wl->value);
       old=temp_wl;
       temp_wl=temp_wl->next;
       if (is_file_traced(ii->file,ii->fn,ii->line) && ii->line!=start_ii->line) {
	  *new_wl=insert_node_into_list(ii,inst,*new_wl);
       } else {
	  for (i=0;i<ii->use_port;i++) {
              vs=ii->uports[i];
              if (vs) {
	         //printf("\t\t\tvs has %d elements\n",vs->used);
                 if (vs->used>inst && vs->iis[inst]) {
                       tmp_ii=vs->iis[inst];
	               tmp_inst=vs->insts[inst];
			
                       if (tmp_inst<tmp_ii->cur_inst && get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\t\t\ttake edge <%x,%s,%s,%d, %d> to <%x,%s,%s,%d,%d\n",ii->orig_addr, ii->file, ii->fn, ii->line, inst, tmp_ii->orig_addr, tmp_ii->file, tmp_ii->fn, tmp_ii->line, vs->insts[inst]);
#endif
    	                 reset_vs_visited_bit(tmp_ii,tmp_inst);
    	                 //set_vs_visited_bit(tmp_ii,tmp_inst);
                         temp_wl=insert_node_into_list(tmp_ii, tmp_inst, temp_wl);
		       }//if
                 }
              }//if(vs
          }//for
       }
       free(old);
    }//while
 }

//filter the predicates, breadth first search
slice_node * slice_from_breadth_first(slice_node * work_list,int * step)
//slice_node * filter_predicate(slice_node * work_list,int * step)
{
     slice_node * result=NULL,* new_wl=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;
     instr_info * tmp_ii;
     Int tmp_inst;
     *step=1;
     while (work_list) {
	//printf("step %d\n",*step);
        while (work_list) {
           ii=work_list->ii;
           inst=work_list->inst;
#undef DEBUG
#ifdef DEBUG
           printf("\tprocess %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value & 0x7f);
#endif
	   old=work_list;
           work_list=work_list->next;
           if (is_file_traced(ii->file,ii->fn, ii->line)) {
               //temp=create_slice_node(ii,inst);
               old->next=NULL;
               result=insert_node_into_list_with_value_step(ii,inst,old->value,(*step), result);
               //work_list=work_list->next;
               //continue;
           } 
	   //slice_one_step_depth_first(old, &new_wl);
            
	   //work_list=work_list->next;

           for (i=0;i<ii->use_port;i++) {
              vs=ii->uports[i];
              if (vs) {
#ifdef DEBUG
	         printf("\t\t\tvs has %d elements\n",vs->used);
#endif
                 if (vs->used>inst && vs->iis[inst] ) {
		    tmp_ii=vs->iis[inst];
	            tmp_inst=vs->insts[inst];
                    

                    if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                       printf("\t\t\tinsert to wl %d(%x,%s,%s,%d) %d\n", tmp_ii->id, tmp_ii->orig_addr,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	               set_vs_visited_bit(tmp_ii,tmp_inst);
                       new_wl=insert_node_into_list(tmp_ii, tmp_inst, new_wl);
		    }//if
                 }//if
              }//if(vs)
            }//for
            

	    free(old);
         }//second while
#ifdef DEBUG
	 printf("NEW WORKLIST:\n");
         print_slice_list(new_wl);
#endif
	 work_list=new_wl;
	 //if (is_worklist_contain_error(work_list)) return result;
	 new_wl=NULL;
         (*step)++;	 
     }
     return result;
}


	  
//starting from the output, reach the cut whose files are traced
slice_node * slice_from_breadth_first_old(slice_node * work_list,int * step)
{
     slice_node * result=NULL,* new_wl=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;
     instr_info * tmp_ii;
     Int tmp_inst;
     *step=1;
     while (work_list) {
	printf("step %d\n",*step);
        while (work_list) {
           ii=work_list->ii;
           inst=work_list->inst;
#undef DEBUG
#ifdef DEBUG
           printf("\tprocess %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
	   old=work_list;
           work_list=work_list->next;
           if (is_file_traced(ii->file,ii->fn, ii->line)) {
               //temp=create_slice_node(ii,inst);
               old->next=NULL;
               result=insert_node_into_list(ii,inst, result);
               //work_list=work_list->next;
               //continue;
           } 
	   //slice_one_step_depth_first(old, &new_wl);
            
	   //work_list=work_list->next;

           for (i=0;i<ii->use_port;i++) {
              vs=ii->uports[i];
              if (vs) {
#ifdef DEBUG
	         printf("\t\t\tvs has %d elements\n",vs->used);
#endif
                 if (vs->used>inst && vs->iis[inst] ) {
		    tmp_ii=vs->iis[inst];
	            tmp_inst=vs->insts[inst];
                   
                    if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                       printf("\t\t\tinsert to wl %d(%x,%s,%s,%d) %d\n", tmp_ii->id, tmp_ii->orig_addr,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	               set_vs_visited_bit(tmp_ii,tmp_inst);
                       new_wl=insert_node_into_list(tmp_ii, tmp_inst, new_wl);
		    }//if
                 }//if
              }//if(vs)
            }//for
            

	    free(old);
         }//second while
#ifdef DEBUG
	 printf("NEW WORKLIST:\n");
         print_slice_list(new_wl);
#endif
	 work_list=new_wl;
	 //if (is_worklist_contain_error(work_list)) return result;
	 new_wl=NULL;
         (*step)++;	 
     }
     return result;
}


slice_node * data_slice_breadth_first(slice_node * work_list,int * step)
//slice_node * filter_predicate(slice_node * work_list,int * step)
{
     slice_node * result=NULL,* new_wl=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;
     instr_info * tmp_ii;
     Int tmp_inst;
     *step=1;
     while (work_list) {
	//printf("step %d\n",*step);
        while (work_list) {
           ii=work_list->ii;
           inst=work_list->inst;
//#define DEBUG
//#undef DEBUG
#ifdef DEBUG
           printf("\tprocess %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value & 0x7f);
#endif
	   old=work_list;
           work_list=work_list->next;
           if (!strncmp (ii->file,"rfc822.c",8)) {
                printf("\t\t\tRFC822.C,%s,%d is %d\n",ii->fn,ii->line,is_file_traced(ii->file,ii->fn,ii->line));
           }

           if (is_file_traced(ii->file,ii->fn, ii->line)) {
               //temp=create_slice_node(ii,inst);
               old->next=NULL;
               result=insert_node_into_list_with_value_step(ii,inst,old->value,(*step), result);
               //work_list=work_list->next;
               //continue;
           } 
	   //slice_one_step_depth_first(old, &new_wl);
            
	   //work_list=work_list->next;

           for (i=1;i<ii->use_port;i++) {
              vs=ii->uports[i];
              if (vs) {
#ifdef DEBUG
	         printf("\t\t\tvs has %d elements\n",vs->used);
#endif
                 if (vs->used>inst && vs->iis[inst] ) {
		    //printf("get tmp ii from %d %d\n", inst, vs->used);
		    tmp_ii=vs->iis[inst];
	            tmp_inst=vs->insts[inst];
		    if (tmp_inst<0) continue;
                    

                    if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                       printf("\t\t\tinsert to wl %d(%x,%s,%s,%d) %d\n", tmp_ii->id, tmp_ii->orig_addr,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	               set_vs_visited_bit(tmp_ii,tmp_inst);
                       new_wl=insert_node_into_list(tmp_ii, tmp_inst, new_wl);
		    }//if
                 }//if
              }//if(vs)
            }//for
	    free(old);
         }//second while
#ifdef DEBUG
	 printf("NEW WORKLIST:\n");
         print_slice_list(new_wl);
#endif
	 work_list=new_wl;
	 //if (is_worklist_contain_error(work_list)) return result;
	 new_wl=NULL;
         (*step)++;	 
     }
     return result;
}


//starting from the output, reach the cut whose files are traced
slice_node * data_slice(slice_node * work_list)
{
     slice_node * result=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;

     instr_info * tmp_ii;
     Int tmp_inst;
     Int traversed=0;

     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
        traversed++;
        if (traversed % 1000==0) { 
           //printf("TRAVERSED %d\n",traversed);
        }
//#undef DEBUG
#define DEBUG
#ifdef DEBUG
        printf("process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
	old=work_list;
        work_list=work_list->next;
        if (is_file_traced(ii->file,ii->fn,ii->line)) {
            //temp=create_slice_node(ii,inst);
            old->next=NULL;
            result=insert_node_into_list(ii,inst, result);
            //work_list=work_list->next;
            //continue;
        } 
	//work_list=work_list->next;
#ifdef DEBUG
        //printf("\t use_port=%d\n",ii->use_port);
#endif
        for (i=1;i<ii->use_port;i++) {
            vs=ii->uports[i];
            if (vs) {
#ifdef DEBUG
	       //printf("\tvs has %d elements\n",vs->used);
#endif
               if (vs->used>inst && vs->iis[inst]) {
		  tmp_ii=vs->iis[inst];
	          tmp_inst=vs->insts[inst];
                  if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\tinsert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	                 set_vs_visited_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
		  }//if
               }
            }
         }//for
	free(old);
#ifdef DEBUG
	//printf("WORKLIST:\n");
        //print_slice_list(work_list);
#endif
     }
     return result;
#undef DEBUG
}

//get the immediate CD for work_list
slice_node * get_immediate_cd(slice_node * work_list)
{
     slice_node * result=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;

     instr_info * tmp_ii;
     Int tmp_inst;
     Int traversed=0;

     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
#undef DEBUG
//#define DEBUG
#ifdef DEBUG
        printf("process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
        work_list=work_list->next;
#ifdef DEBUG
        //printf("\t use_port=%d\n",ii->use_port);
#endif
            vs=ii->uports[0];
            if (vs) {
#ifdef DEBUG
	       //printf("\tvs has %d elements\n",vs->used);
#endif
               if (vs->used>inst && vs->iis[inst]) {
		  tmp_ii=vs->iis[inst];
	          tmp_inst=vs->insts[inst];
                  if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\tinsert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	                 set_vs_visited_bit(tmp_ii,tmp_inst);
                         result=insert_node_into_list(tmp_ii, tmp_inst, result);
		  }//if
               }
            }
#ifdef DEBUG
	//printf("WORKLIST:\n");
        //print_slice_list(work_list);
#endif
     }
     return result;
#undef DEBUG
}



//starting from the output, reach the cut whose files are traced
slice_node * slice_from_traced_cut(slice_node * work_list)
{
     slice_node * result=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;

     instr_info * tmp_ii;
     Int tmp_inst;
     Int traversed=0;

     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
        traversed++;
        if (traversed % 1000==0) {
	   //if (traversed>10000) exit(-1); 
           //printf("TRAVERSED %d\n",traversed);
        }
#undef DEBUG
//#define DEBUG
#ifdef DEBUG
        printf("process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
	old=work_list;
        work_list=work_list->next;
        if (is_file_traced(ii->file,ii->fn,ii->line)) {
            //temp=create_slice_node(ii,inst);
            old->next=NULL;
            result=insert_node_into_list(ii,inst, result);
            //work_list=work_list->next;
            //continue;
        }
	if (ii->line==129) printf("XXX 129, is_file_traced=%d\n",is_file_traced(ii->file,ii->fn,ii->line)); 
	//work_list=work_list->next;
#ifdef DEBUG
        //printf("\t use_port=%d\n",ii->use_port);
#endif
        for (i=0;i<ii->use_port;i++) {
            vs=ii->uports[i];
	       if (ii->id==3173) {
		   printf("3173 good 1\n");
               }
            if (vs) {
#ifdef DEBUG
	       //printf("\tvs has %d elements\n",vs->used);
#endif
	       if (ii->id==3173) {
		   printf("3173 good 2\n");
               }
               if (vs->used>inst && vs->iis[inst]) {
	          if (ii->id==3173) {
		     printf("3173 good 3\n");
                  }

		  tmp_ii=vs->iis[inst];
	          tmp_inst=vs->insts[inst];
                  if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\tinsert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	                 set_vs_visited_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
		  }//if
               }
            }
         }//for
	free(old);
#ifdef DEBUG
	//printf("WORKLIST:\n");
        //print_slice_list(work_list);
#endif
     }
     return result;
#undef DEBUG
}


slice_node * error_slice_breadth_first(slice_node * work_list)
{
     slice_node * result=NULL;
     slice_node * new_wl=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;

     instr_info * tmp_ii;
     Int tmp_inst;
     Int traversed=0;
     potential_dep * pd;
     Int step=0;

  while (work_list) {
     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
        traversed++;
        if (traversed % 1000==0) { 
           printf("TRAVERSED %d\n",traversed);
        }
#define DEBUG
//#undef DEBUG
#ifdef DEBUG
        printf("E process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
	old=work_list;
        work_list=work_list->next;
        if (is_file_traced(ii->file,ii->fn,ii->line)) {
            //temp=create_slice_node(ii,inst);
            old->next=NULL;
            result=insert_node_into_list_with_value_step(ii,inst, old->value, step, result);
            //work_list=work_list->next;
            //continue;
        } 
	//work_list=work_list->next;
#ifdef DEBUG
        //printf("\t use_port=%d\n",ii->use_port);
#endif
        for (i=0;i<ii->use_port;i++) {
            vs=ii->uports[i];
            if (vs) {
#ifdef DEBUG
	       //printf("\tvs has %d elements\n",vs->used);
#endif
               if (vs->used>inst && vs->iis[inst]) {
		  tmp_ii=vs->iis[inst];
	          tmp_inst=vs->insts[inst];
                  //if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {

                  //if (!get_vs_visited_bit(tmp_ii,tmp_inst)&&(!get_certified_bit(tmp_ii,tmp_inst)||i==0)) {
                  if (!get_vs_visited_bit(tmp_ii,tmp_inst)&&(!get_certified_bit(tmp_ii,tmp_inst))) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\tinsert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	                 set_vs_visited_bit(tmp_ii,tmp_inst);
                         new_wl=insert_node_into_list(tmp_ii, tmp_inst, new_wl);
		  }//if
               }
            }
         }//for

#ifdef CONSIDER_POTENTIAL
         if (ii->pds && ii->pds_alloced > inst) {
             pd=ii->pds[inst];
  	     //if (ii->line==466 && inst==44) pd=NULL;
  	     //if (ii->line==439 || ii->line==440 || ii->line==390) pd=NULL;
             while (pd) {
                tmp_ii=pd->pred_ii;
                tmp_inst=pd->pred_inst;
		//if (tmp_ii->line==439|| tmp_ii->line==440) { pd=pd->next; continue;}
                //if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);             
                if (!get_vs_visited_bit(tmp_ii,tmp_inst)&&(!get_certified_bit(tmp_ii,tmp_inst))) {
#ifdef DEBUG
                         printf("\tPD insert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
                         set_vs_visited_bit(tmp_ii,tmp_inst);
                         new_wl=insert_node_into_list(tmp_ii, tmp_inst, new_wl);
                }//if
                pd=pd->next;
             }//while

          }//if (potential dep)
#endif

	free(old);
      }//second while
#ifdef DEBUG
	//printf("WORKLIST:\n");
        //print_slice_list(work_list);
#endif
	work_list=new_wl;
	new_wl=NULL;
        step++;
     }
     return result;
#undef DEBUG
}


//starting from the erroneous output, stop populated at certified instances
slice_node * error_slice(slice_node * work_list)
{
     slice_node * result=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;

     instr_info * tmp_ii;
     Int tmp_inst;
     Int traversed=0;
     potential_dep * pd;

     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
        traversed++;
        if (traversed % 1000==0) { 
           printf("TRAVERSED %d\n",traversed);
        }
#define DEBUG
//#undef DEBUG
#ifdef DEBUG
        printf("E process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
	old=work_list;
        work_list=work_list->next;
        if (is_file_traced(ii->file,ii->fn,ii->line)) {
            //temp=create_slice_node(ii,inst);
            old->next=NULL;
            result=insert_node_into_list(ii,inst, result);
            //work_list=work_list->next;
            //continue;
        } 
	//work_list=work_list->next;
#ifdef DEBUG
        //printf("\t use_port=%d\n",ii->use_port);
#endif
        for (i=0;i<ii->use_port;i++) {
            vs=ii->uports[i];
            if (vs) {
#ifdef DEBUG
	       //printf("\tvs has %d elements\n",vs->used);
#endif
               if (vs->used>inst && vs->iis[inst]) {
		  tmp_ii=vs->iis[inst];
	          tmp_inst=vs->insts[inst];
                  //if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {

                  //if (!get_vs_visited_bit(tmp_ii,tmp_inst)&&(!get_certified_bit(tmp_ii,tmp_inst)||i==0)) {
                  if (!get_vs_visited_bit(tmp_ii,tmp_inst)&&(!get_certified_bit(tmp_ii,tmp_inst))) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\tinsert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	                 set_vs_visited_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
		  }//if
               }
            }
         }//for

#ifdef CONSIDER_POTENTIAL
         if (ii->pds && ii->pds_alloced > inst) {
             pd=ii->pds[inst];
  	     //if (ii->line==466 && inst==44) pd=NULL;
  	     //if (ii->line==439 || ii->line==440 || ii->line==390) pd=NULL;
             while (pd) {
                tmp_ii=pd->pred_ii;
                tmp_inst=pd->pred_inst;
		//if (tmp_ii->line==439|| tmp_ii->line==440) { pd=pd->next; continue;}
                //if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);             
                if (!get_vs_visited_bit(tmp_ii,tmp_inst)&&(!get_certified_bit(tmp_ii,tmp_inst))) {
#ifdef DEBUG
                         printf("\tPD insert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
                         set_vs_visited_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
                }//if
                pd=pd->next;
             }//while
          }//if (potential dep)
#endif

	free(old);
#ifdef DEBUG
	//printf("WORKLIST:\n");
        //print_slice_list(work_list);
#endif
     }
     return result;
#undef DEBUG
}

int threshold=1;

void set_threshold(Int t)
{
   threshold=t;
}

//The candidate set of certified nodes are marked as visited, in other words, first do BS on all output including correct and erroneous outputs.
slice_node * certify_slice(slice_node * work_list, Bool dice)
{
     slice_node * result=NULL;
     instr_info * ii=NULL;
     forward_edge * edge;
     Int inst,i,value;
     slice_node * temp, * old;
     potential_dep * pd;

     instr_info * cd_iis[256];
     Int cd_insts[256];
     Int cd_idx;
     

     instr_info * tmp_ii, * cd_son_ii;
     Int tmp_inst, cd_son_inst;
     Int certified,visited;
     Int traversed=0;
     val_stream * vs;
     Int edge_cnt=0;
     Bool do_it;
     printf("threshold=%d\n",threshold); 
     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
        traversed++; 
        if (traversed % 1000==0) {
           //printf("TRAVERSED %d\n",traversed);
        }
#define DEBUG
//#undef DEBUG
#ifdef DEBUG
        printf("C process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
        old=work_list;
        work_list=work_list->next;
        if (is_file_traced(ii->file,ii->fn,ii->line)) {
            //temp=create_slice_node(ii,inst);
            old->next=NULL;
            result=insert_node_into_list(ii,inst, result);
            //work_list=work_list->next;
            //continue;
        }
        //certificates are propagaged along data dependence edges
	do_it=True;
/*
        if (ii->is_predicate && ii->allocated>inst) {
	    value=ii->values[inst];
	    if (value && !ii->true_propagate) do_it=False;
	    if (!value && !ii->false_propagate) do_it=False;
	}
*/
#ifdef MAPPING
	//if (ii->allocated>inst && ii->is_predicate ) {
	//if (!dice && ii->allocated>inst && get_vs_visited_bit(ii,inst)) {
	//if (!dice && ii->allocated>inst && ii->is_predicate) {
	if (!dice && ii->allocated>inst && ii->is_predicate && get_vs_visited_bit(ii,inst)) {
	    value=ii->values[inst]; 
	    //if (get_mapping(ii,value)>100) do_it=False;
	    if (get_mapping(ii,value)>threshold) do_it=False;
	    else if (ii->allocated ==threshold && (value==0 || value==1) ) do_it=False;
	   /* 
	   else {
		//Int tt=0;
		//for (i=0;i<ii->allocated;i++) if (ii->values[i]==value) tt++; 
		//if (tt*threshold <= ii->allocated && (value==0 || value==1) ) do_it=False;
		//if (tt >=threshold && (value==0 || value==1) ) do_it=False;
	    }
	  */
	}
	//if (ii->line==511 && inst==8) continue;
#endif

    if (do_it) {
        for (i=1;i<ii->use_port;i++) {
            vs=ii->uports[i];
            if (vs) {
#ifdef DEBUG
	       //printf("\tvs has %d elements\n",vs->used);
#endif
               if (vs->used>inst && vs->iis[inst]) {
		  tmp_ii=vs->iis[inst];
	          tmp_inst=vs->insts[inst];
                  if (!get_certified_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\tinsert to wl %d(%x,%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->orig_addr,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	                 set_certified_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
		  }//if
               }
            }
         }//for


#ifdef CONSIDER_POTENTIAL
         if (ii->pds && ii->pds_alloced > inst) {
             pd=ii->pds[inst];
	     
  	     //if (ii->line==466 && inst==44) pd=NULL;
  	     //if (ii->id==4121 && inst==90) pd=NULL;
  	     //if (ii->line==439|| ii->line==440 || ii->line==390) pd=NULL;
             while (pd) {
                tmp_ii=pd->pred_ii;
                tmp_inst=pd->pred_inst;
		
		//if (tmp_ii->line==439 || tmp_ii->line==440) { pd=pd->next; continue;}
		
		if ((ii->line==506 && tmp_ii->line==503)
		    || ii->line==503
		    || ii->line==466
		    || ii->line==439
		    || ii->line==215
		    || ii->line==526
		    || ii->line==460
		    || ii->line==244
		    || (ii->line==396 && tmp_ii->line==439)
		    || (ii->line==396 && tmp_ii->line==440)
		    || (ii->line==390 && tmp_ii->line==440)
		    || (ii->line==390 && tmp_ii->line==439)
		    ) {
			pd=pd->next;
			continue;
		}
		
                if (!get_certified_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);             
#ifdef DEBUG
                         printf("\tPD insert %d(%x,%s,%d) %d to wl %d(%x,%s,%d) %d\n",ii->id, ii->orig_addr, ii->file, ii->line, inst, tmp_ii->id, tmp_ii->orig_addr, tmp_ii->file, tmp_ii->line, tmp_inst);
#endif
                         set_certified_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
                }//if
                pd=pd->next;
             }//while
          }//if (potential dep)
#endif



    }

    cd_idx=0;
    vs=ii->uports[0];

    if (vs) {
        if (vs->used>inst && vs->iis[inst]) {
	    cd_iis[cd_idx] = vs->iis[inst];
	    cd_insts[cd_idx++]=vs->insts[inst];
        }
    }
/*
#ifdef CONSIDER_POTENTIAL
    if (ii->pds && ii->pds_alloced > inst) {
        pd=ii->pds[inst];
        while (pd) {
	   cd_iis[cd_idx]=pd->pred_ii;
	   cd_insts[cd_idx++]=pd->pred_inst;
	   pd=pd->next;
        }
    }

#endif
*/
     
    //for control depdendence edges, it requires all the control depdendent children of the predicate are certified.
         //vs=ii->uports[0];
	 //if (vs) {
             //if (vs->used>inst && vs->iis[inst]) {
                //tmp_ii=vs->iis[inst];
                //tmp_inst=vs->insts[inst];
       for (i=0;i<cd_idx;i++) {
	   tmp_ii=cd_iis[i];
	   tmp_inst=cd_insts[i];

#ifdef DEBUG
                printf("\t Try CD <%d, %x,%s,%s,%d> %d\n", tmp_ii->id, tmp_ii->orig_addr, tmp_ii->file, tmp_ii->fn, tmp_ii->line, tmp_inst);   
#endif
		if (!get_certified_bit(tmp_ii,tmp_inst)) {
	  	  //certified=1;
		  certified=0;
		  edge_cnt=0;
		  visited=0;
		  if (tmp_inst<tmp_ii->forward_inst) {
			edge=tmp_ii->edges[tmp_inst];
		        while (edge) {
			   
			   cd_son_ii=edge->ii;
			   cd_son_inst=edge->inst;
#ifdef DEBUG
			   printf("\t\ttry son <%d, %x,%s,%s,%d> %d C=%d\n", cd_son_ii->id, cd_son_ii->orig_addr, cd_son_ii->file, cd_son_ii->fn, cd_son_ii->line, cd_son_inst,get_certified_bit(cd_son_ii,cd_son_inst));
#endif


//OPTITON ONE is take all the reachable fro output as candidate
#ifdef OPTION_ONE
			   //if it is a candidate but not certified
			   if (get_vs_visited_bit(cd_son_ii,cd_son_inst)) { 
                               if (!get_certified_bit(cd_son_ii,cd_son_inst)		       //&& !cd_son_ii->is_predicate
			          ) {
#ifdef DEBUG
			            printf("\t\t\tfailed\n");
#endif
			       //certified=0;
			       //break;
			       } else certified++;
			       visited++;

                           }			   
                           //if (get_certified_bit(cd_son_ii,cd_son_inst)) certified++;
#endif

#define OPTION_TWO
#ifdef OPTION_TWO
			   if (get_certified_bit(cd_son_ii,cd_son_inst)) {
			      certified++;
			      //break;
			   } 
#endif




			   edge=edge->next;
			   edge_cnt++;
	
			}//while
		   } else {//to be greedy
			//there is no forward edge
		   }

              

		   //if (2*certified>=visited) {
		   //if (2*certified>=edge_cnt) {
		   if (certified) {
#ifdef DEBUG
			 printf("\t CD certified\n");
#endif
	    	         set_certified_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
		   }//if cerfified
                }//if (!get
	   }//for
         //}
	 //printf("\t CD done\n");
	 free(old);
     }//whlie
     return result;
}

       
void step_debug_certify_slice()
{
     forward_edge * edge;
     instr_info * ii,*tmp_ii;
     Int id,inst,tmp_inst;
     slice_node * wl,*old;
     Int certified;
     //while (1) {
	//scanf("%d,%d",&id,&inst);
	id=1869; inst=1;
	//if (id>=instr_gid) continue;
	ii=iis[id];
	//if (inst>=ii->cur_inst) continue;
	printf("For %d(%x,%s,%s,%d), %d\n",id,ii->orig_addr, ii->file, ii->fn, ii->line, inst);
        wl=NULL;
	wl=insert_node_into_list(ii,inst,wl);
	while (wl) {
	    ii=wl->ii;
	    inst=wl->inst;
            printf("process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,wl->value,wl->value&0x7f);
	    old=wl;
	    wl=wl->next;
	    if (inst<ii->forward_inst) {
   	      edge=ii->edges[inst];
	      certified=0;
	      while (edge) {
	        tmp_ii=edge->ii;
	        tmp_inst=edge->inst;
                edge=edge->next;
	        if (!get_vs_visited_bit(tmp_ii,tmp_inst)) continue;
	   	printf("\tson: %d(%x,%s,%s,%d) %d C=%d P=%d\n",tmp_ii->id, tmp_ii->orig_addr, tmp_ii->file, tmp_ii->fn, tmp_ii->line, tmp_inst, get_certified_bit(tmp_ii,tmp_inst),tmp_ii->is_predicate);
                if (get_certified_bit(tmp_ii,tmp_inst)) certified=1;
              }//while
	      if (!certified) {
		edge=ii->edges[inst];
	        while (edge) {
	            tmp_ii=edge->ii;
	            tmp_inst=edge->inst;
                    edge=edge->next;
	            //if (get_forward_visited_bit(tmp_ii,tmp_inst)) continue;
	            if (!get_vs_visited_bit(tmp_ii,tmp_inst) || get_forward_visited_bit(tmp_ii,tmp_inst)) continue;
		    wl=insert_node_into_list(tmp_ii,tmp_inst,wl);
		    set_forward_visited_bit(tmp_ii,tmp_inst);
		}
	    }//if(!cer...
	    free(old);
	  }	
	}//while (wl)
    //}//while(1)
}
		 
slice_node * forward_slice_bf(slice_node * work_list)
{
     slice_node * result=NULL,*new_wl;
     instr_info * ii=NULL;
     forward_edge * edge;
     Int inst,i;
     slice_node * temp, * old;

     instr_info * tmp_ii;
     Int tmp_inst;
     Int traversed=0;
     Int step=0;
     //Int step=0xffff;

     while (work_list) {
	new_wl=NULL;
        while (work_list) {
           ii=work_list->ii;
           inst=work_list->inst;
           traversed++;
           if (traversed % 1000==0) { 
              //printf("TRAVERSED %d\n",traversed);
           }
#undef DEBUG
//#define DEBUG
#ifdef DEBUG
           printf("process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
	   old=work_list;
           work_list=work_list->next;
           if (is_file_traced(ii->file,ii->fn,ii->line)) {
               //temp=create_slice_node(ii,inst);
               old->next=NULL;
               //result=insert_node_into_list(ii,inst, result);
               result=insert_node_into_list_with_value_step(ii,inst,old->value,step, result);
               //work_list=work_list->next;
               //continue;
           }
           if (inst<ii->forward_inst) { 
 	      //work_list=work_list->next;
              edge=ii->edges[inst];
	      while (edge) {
	          tmp_ii=edge->ii;
                  tmp_inst=edge->inst;
                  if (!get_forward_visited_bit(tmp_ii,tmp_inst)) {
#ifdef DEBUG
                      printf("\tinsert to wl %d(%x,%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->orig_addr, tmp_ii->file, tmp_ii->fn, tmp_ii->line, tmp_inst);
#endif
    	              set_forward_visited_bit(tmp_ii,tmp_inst);
                      new_wl=insert_node_into_list(tmp_ii, tmp_inst, new_wl);
                  }//if
	          edge=edge->next;
               }//while
	    }//if (inst<
        
	    free(old);
	 }//while(work_list)
	 step++;
	 //step--;
	 work_list=new_wl;
#ifdef DEBUG
	//printf("WORKLIST:\n");
        //print_slice_list(work_list);
#endif
     }
     return result;
#undef DEBUG
}

slice_node * forward_slice(slice_node * work_list)
{
     slice_node * result=NULL;
     instr_info * ii=NULL;
     forward_edge * edge;
     Int inst,i;
     slice_node * temp, * old;

     instr_info * tmp_ii;
     Int tmp_inst;
     Int traversed=0;

     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
        traversed++;
        if (traversed % 1000==0) { 
           //printf("TRAVERSED %d\n",traversed);
        }
#define DEBUG
#ifdef DEBUG
        printf("F process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value& 0x7f);
#endif
	old=work_list;
        work_list=work_list->next;
        if (is_file_traced(ii->file,ii->fn,ii->line)) {
            //temp=create_slice_node(ii,inst);
            old->next=NULL;
            result=insert_node_into_list(ii,inst, result);
            //work_list=work_list->next;
            //continue;
        }
        if (inst<ii->forward_inst) { 
 	   //work_list=work_list->next;
           edge=ii->edges[inst];
	   while (edge) {
	       tmp_ii=edge->ii;
               tmp_inst=edge->inst;

               if (!get_forward_visited_bit(tmp_ii,tmp_inst)) {
#ifdef DEBUG
                   printf("\tinsert to wl %d(%x,%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->orig_addr, tmp_ii->file, tmp_ii->fn, tmp_ii->line, tmp_inst);
#endif
    	           set_forward_visited_bit(tmp_ii,tmp_inst);
                   work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
               }//if
	       edge=edge->next;
            }//while
        }
	free(old);
#ifdef DEBUG
	//printf("WORKLIST:\n");
        //print_slice_list(work_list);
#endif
     }
     return result;
#undef DEBUG
}

//#define DEBUG

//starting from the output, reach the cut whose files are traced
slice_node * slice_to_traced_cut(slice_node * work_list)
{
     slice_node * result=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;
     instr_info * tmp_ii;
     Int tmp_inst;
     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
#define DEBUG
#ifdef DEBUG
        printf("process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
	old=work_list;
        work_list=work_list->next;
        if (is_file_traced(ii->file,ii->fn,ii->line)) {
            //temp=create_slice_node(ii,inst);
            old->next=NULL;
            result=insert_node_into_list(ii,inst, result);
            //work_list=work_list->next;
            //continue;
        } else {
	   //work_list=work_list->next;
#ifdef DEBUG      
           //printf("\t use_port=%d\n",ii->use_port);
#endif
           for (i=0;i<ii->use_port;i++) {
               vs=ii->uports[i];
               if (vs) {
#ifdef DEBUG
		   //printf("\tvs has %d elements,inst=%d, vs->iis[inst]=%x\n",vs->used,inst, vs->iis[inst]);
#endif
                   
                   if (vs->used>inst && vs->iis[inst]) {
                       tmp_ii=vs->iis[inst];
	               tmp_inst=vs->insts[inst];
                       if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\tinsert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	                 set_vs_visited_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
		       }//if

                   }
               }
            }//for
            //DEBUG
	    //result=insert_node_into_list(ii, inst, result);

        }
	free(old);
#ifdef DEBUG
	//printf("WORKLIST:\n");
        //print_slice_list(work_list);
#endif
     }
     return result;
}

int  print_slice_list_condensed_w_search(slice_node * head, Int error, Bool * error_in)
{
     slice_node * list=head;
     Char * old_file=NULL;
     Int old_line=-1;
     Int cnt=0;
     Int size=0;
     * error_in=False;
     while (list) {
	if (old_line!=list->ii->line || (old_file !=NULL && strcmp(old_file, list->ii->file))) {
             size++;
             printf("%d:< %d(%x, %s, %s, %d) >\n", cnt++,list->ii->id,list->ii->orig_addr, list->ii->file, list->ii->fn, list->ii->line);
	     old_line=list->ii->line;
	     old_file=list->ii->file;
	     if (old_line==error)  *error_in=True;
        }
        list=list->next;
     }
     printf("CONDENSED SIZE is %d\n",size);
     return size;
}


int  print_slice_list_condensed(slice_node * head)
{
     slice_node * list=head;
     Char * old_file=NULL;
     Int old_line=-1;
     Int cnt=0;
     Int size=0;
     while (list) {
	if (old_line!=list->ii->line || (old_file !=NULL && strcmp(old_file, list->ii->file))) {
             size++;
             printf("%d:< %d(%x, %s, %s, %d) >\n", cnt++,list->ii->id,list->ii->orig_addr, list->ii->file, list->ii->fn, list->ii->line);
	     old_line=list->ii->line;
	     old_file=list->ii->file;
        }
        list=list->next;
     }
     printf("CONDENSED SIZE is %d\n",size);
     return size;
}

Int stat_slice_list(slice_node * head)
{
     slice_node * list=head;
     Int size=0;
     Int predicate=0;
     while (list) {
	size++;
	if (list->ii->is_predicate) predicate++;
        //printf("< %d(%x, %s, %s, %d) , %d ,v=%x '%c'>\n", list->ii->id,list->ii->orig_addr, list->ii->file, list->ii->fn, list->ii->line, list->inst,list->value,list->value);
        list=list->next;
     }
     //printf("DONE printing size=%d(predicate=%d)\n",size,predicate);
     predicate_size=predicate;
     return size;
}

Int print_predicate_list(slice_node * head)
{
     slice_node * list=head;
     Int size=0;
     Int predicate=0;
     while (list) {
	if (list->ii->jmp_pc) {
	   size++;
           printf("%x %d %d %d %d<%s,%s,%d>\n", list->ii->jmp_pc,list->inst,list->step, list->value,list->ii->id,list->ii->file, list->ii->fn,list->ii->line);
        }
	list=list->next;
     }
     printf("DONE printing size=%d\n",size);
     return size;
}

Int print_slice_list(slice_node * head)
{
     slice_node * list=head;
     Int size=0,c;
     Int predicate=0;
     instr_info * ii;
     while (list) {
	size++;
	if (list->ii->is_predicate) predicate++;
        ii=list->ii;
        if (ii->input_related&&ii->input_related[list->inst]!=0xfffe) { 
           printf("%d < %d(%x, %s, %s, %d) , %d ,v=%x '%c'> INPUT='%c', step=%d\n", size, ii->id,ii->orig_addr, ii->file, ii->fn, ii->line, list->inst,list->value,list->value&0x7f,ii->input_related[list->inst],list->step);
        } else {
	   c=list->value;
	   if (list->value<5 || list->value>128) c='X';
	   printf("%d < %d(%x, %s, %s, %d) , %d ,v=%x '%c' step=%d>\n", size,ii->id,ii->orig_addr, ii->file, ii->fn, ii->line, list->inst,list->value,c,list->step);
        }
        list=list->next;
     }
     printf("DONE printing size=%d(predicate=%d)\n",size,predicate);
     predicate_size=predicate;
     return size;
}

void write_predicate_list(Char *filename,slice_node * head)
{
     slice_node * list=head;
     Int size=0;
     Int predicate=0;
     FILE * fd=fopen(filename, "w");
     printf("DUMPING...\n");
     while (list) {
	if (list->ii->jmp_pc) {
	   size++;
           fprintf(fd,"%x %d %d %d %d<%s,%s,%d>\n", list->ii->jmp_pc,list->inst,list->step, list->value,list->ii->id,list->ii->file,list->ii->fn,list->ii->line);
        }
	list=list->next;
     }
     fclose(fd);
}

void write_slice_list(Char * filename,slice_node * head)
{
     FILE * fd=fopen(filename, "w");
     slice_node * list=head;
     printf("DUMPING...\n");
     while (list) {
        fprintf(fd,"<%d(%x, %s, %s, %d) , %d ,v=%x >\n", list->ii->id,list->ii->orig_addr, list->ii->file, list->ii->fn, list->ii->line, list->inst,list->value);
        list=list->next;
     }
     fclose(fd);
}

slice_node * read_slice_list(Char * filename)
{
    FILE * fd=fopen(filename,"r");
    slice_node * head=NULL,*temp,*tail=NULL;
    Char buf[1024];
    Int i;
    instr_info * ii;
    Int inst, id, orig_addr, line,start;
    Char * file, * fn;
    UInt v;

    while (fgets(buf,1024,fd)) {
       i=1;
       start=i;
       while (buf[i]!='(') i++;
       buf[i]='\0';
       id=atoi(buf+start);
       //printf("%d\n",id);
       
       i++;
       start=i;
       while (buf[i]!=',') i++;
       buf[i]='\0';
       orig_addr=strtol(buf+start, NULL, 16);

       i+=2;
       start=i;
       while (buf[i]!=',') i++;
       buf[i]='\0';
       file=buf+start;

       i+=2;
       start=i;
       while (buf[i]!=',') i++;
       buf[i]='\0';
       fn=buf+start;

       i+=2;
       start=i;
       while (buf[i]!=')') i++;
       buf[i]='\0';
       line=atoi(buf+start);

       i+=4;
       start=i;
       while (buf[i]!=' ') i++;
       buf[i]='\0';
       inst=atoi(buf+start);

       i+=4;
       start=i;
       while (buf[i]!=' ') i++;
       buf[i]='\0';
       v=strtol(buf+start, NULL, 16);
       if (v==10) fgets(buf, 1024,fd);

       if (iis[id]) {
            if (iis[id]->orig_addr!=orig_addr) {
		printf("ERROR, instr_info inconsistent\n");
		exit(-1);
            }
       } else {
	    iis[id]=ii=create_instr_info(orig_addr, -1, fn, file, line,id, -1,-1);
       }
       
       tail=insert_node_into_list_tail_with_value(iis[id],inst,v,tail);
       if (!head) head=tail;
    }//while
    fclose(fd);
    return head;
}
    
//s1 U s2
slice_node * union_slice(slice_node * s1, slice_node * s2)
{
    slice_node * t1=s1;
    slice_node * t2=s2;
    slice_node * temp;
    slice_node * us=NULL, * head=NULL;
    while (t1 || t2) {
       if (!t1) {
	  us=insert_node_into_list_tail_with_value(t2->ii,t2->inst,t2->value,us);
	  t2=t2->next;
       } else if (!t2) {
	  temp=create_slice_node_with_value(t1->ii,t1->inst,t1->value);
	  us=insert_node_into_list_tail_with_value(t1->ii,t1->inst,t1->value,us);
	  t1=t1->next;
       } else {
	  if (t1->ii->orig_addr > t2->ii->orig_addr || (t1->ii->orig_addr==t2->ii->orig_addr && t1->inst > t2->inst)) {
	      us=insert_node_into_list_tail_with_value(t2->ii,t2->inst,t2->value,us);
	      t2=t2->next;
	  } else if (t1->ii->orig_addr< t2->ii->orig_addr || (t1->ii->orig_addr==t2->ii->orig_addr && t1->inst<t2->inst)) {
	      us=insert_node_into_list_tail_with_value(t1->ii,t1->inst,t1->value,us);
	      t1=t1->next;
          } else {     
	      us=insert_node_into_list_tail_with_value(t1->ii,t1->inst,t1->value,us);
	      t1=t1->next; 
              t2=t2->next;
          }
      }
      if (!head) head=us;
   }//while
   return head;
}

//s1-s2
slice_node * diff_slice(slice_node * s1, slice_node * s2)
{
    slice_node * t1=s1;
    slice_node * t2=s2;
    slice_node * temp,*us=NULL;
    slice_node * diff=NULL, *head=NULL;
    instr_info * ii;
    Int inst;
    while (t1) {
      if (!t2) {
      	  temp=create_slice_node_with_value(t1->ii,t1->inst,t1->value);
	  us=insert_node_into_list_tail_with_value(t1->ii,t1->inst,t1->value,us);
	  if (!head) head=us;
	  t1=t1->next;
	  continue;
      }
      ii=t1->ii;
      inst=t1->inst;
      if (t1&& t2)printf("FROM: t2 -->(%d %d), t1-->(%d %d)\n", t2->ii->id, t2->inst, t1->ii->id, t1->inst);
      while (t2 &&(ii->orig_addr > t2->ii->orig_addr || (ii->orig_addr==t2->ii->orig_addr && inst> t2->inst))) t2=t2->next;
      if (!t2 || ii->orig_addr!=t2->ii->orig_addr || inst!=t2->inst) {
	  us=insert_node_into_list_tail_with_value(t1->ii,t1->inst,t1->value,us);
	  if (!head) head=us;
      }
      t1=t1->next;
      if (t1&& t2)printf("t2 -->(%d %d), t1-->(%d %d)\n", t2->ii->id, t2->inst, t1->ii->id, t1->inst);
   }
   return head;
}      

//s1^s2
slice_node * intersect_slice(slice_node * s1, slice_node *s2)
{
    slice_node * t1=s1;
    slice_node * t2=s2;
    slice_node * temp;
    slice_node * us=NULL,*head=NULL;
    instr_info * ii;
    Int inst;
    while (t1&&t2) {
      ii=t1->ii;
      inst=t1->inst;
      while (t2 &&(ii->orig_addr > t2->ii->orig_addr || (ii->orig_addr==t2->ii->orig_addr && inst> t2->inst))) t2=t2->next;
      if (t2 && ii->orig_addr==t2->ii->orig_addr && inst==t2->inst) {
	  us=insert_node_into_list_tail_with_value_step(t1->ii,t1->inst,t1->value,t1->step,us);
	  if (!head) head=us;
	  t2=t2->next;
      }
      t1=t1->next;
   }
   return head;
}   

void stat_executed_LOC()
{
    Int i;
    Int count=0;
    instr_info * ii;
    slice_node * all=NULL,*head=NULL;
    for (i=0;i<instr_gid;i++) {
	ii=iis[i];
	if (!ii) continue;
	if (!is_file_traced(ii->file,ii->fn,ii->line)) continue;
	//printf("insert %d(%s,%s,%d)\n",ii->id, ii->file,ii->fn,ii->line);
        all=insert_node_into_list_tail_with_value(ii,0,0,all);
	if (!head) head=all;
    }
    all=qsort_list(head);
    printf("EXECUTED:\n");
    print_slice_list_condensed(all);
}
void clear_backward_pdg()
{
    int i,j;
    instr_info * ii;
    val_stream * vs;
    for (i=0;i<instr_gid;i++) {
       ii=iis[i];
       for (j=0;j<ii->use_port;j++) {
          vs=ii->uports[j];
	  if (!vs||!vs->iis) continue;
	  free(vs->iis);
	  free(vs->insts);
	  free(vs);
       }
   }
}

void compute_propagatability()
{
     Int i,j,k,def_inst, val,def_val,true_val,false_val;
     Bool true_available, false_available;
     instr_info * ii, * def_ii;
     val_stream * vs;

     for (i=0;i<instr_gid;i++) {
        ii=iis[i];
	if (!ii) continue;
	if (!ii->is_predicate) continue;
        true_available=false_available=False;
	ii->true_propagate=ii->false_propagate=True;
	printf("comput p for %d\n",ii->id);
        for (j=1;j<ii->use_port;j++) {
          vs=ii->uports[j];
	  if (!vs) continue;
          for (k=0;k<vs->used;k++) {
              def_ii=vs->iis[k];
	      if (!def_ii) continue;
              def_inst=vs->insts[k];
	      if (k>=ii->allocated) continue;
	      val=ii->values[k];
	      if (def_ii->allocated<=def_inst) continue;
	      def_val=def_ii->values[def_inst];
	      if (val) {
		 if (!true_available) {
		     true_available=True;
		     true_val=def_val;
		 } else {//more than one values yield true here
		     if (true_val!=def_val) ii->true_propagate=False;
                 }
              } else {
		 if (!false_available) {
		     false_available=True;
		     false_val=def_val;
		 } else { //more than one values yield false here
		     if (false_val!=def_val) ii->false_propagate=False;
                 }
	      }//if (val)
	      if (!ii->true_propagate && ! ii->false_propagate) break;
           }//for k
	 }//for j
	 printf("< %d(%x, %s, %s, %d)> TRUE=%d, FALSE=%d \n", ii->id,ii->orig_addr, ii->file, ii->fn, ii->line, ii->true_propagate, ii->false_propagate);
     }//for i
}
	
void construct_forward_pdg_dd()
{
    Int i,j,k,def_inst; 
    instr_info * ii, * def_ii;
    val_stream * vs;
    forward_edge * edge;
    potential_dep * pd;

    for (i=0;i<instr_gid;i++) {
        ii=iis[i];
	if (!ii) continue;
        for (j=1;j<ii->use_port;j++) {
          vs=ii->uports[j];
	  if (!vs) continue;
          for (k=0;k<vs->used;k++) {
              def_ii=vs->iis[k];
	      if (!def_ii) continue;
              def_inst=vs->insts[k];
			
	      if (def_inst>=def_ii->forward_inst) def_ii->forward_inst=def_inst+1;
         }
       }
   }
   printf ("Calculated\n");
   for (i=0;i<instr_gid;i++) {
        ii=iis[i];
	if (!ii) continue;
	if (ii->forward_inst < ii->cur_inst) ii->forward_inst=ii->cur_inst;	
        if (ii->forward_inst) {
	   ii->edges=calloc(ii->forward_inst, sizeof(forward_edge*));
           ii->forward_visited=calloc(ii->forward_inst/32+1, sizeof(UInt));
#ifdef DEP_DISTANCE
	   ii->forward_dist=calloc(ii->forward_inst, sizeof(UInt));
#endif     
        } 
		
    }
    printf("ALLOCATED\n"); 
    for (i=0;i<instr_gid;i++) {
        ii=iis[i];
	if (!ii) continue;
        for (j=1;j<ii->use_port;j++) {
          vs=ii->uports[j];
	  if (!vs) continue;
          for (k=0;k<vs->used;k++) {
              def_ii=vs->iis[k];
	      if (!def_ii) continue;
              def_inst=vs->insts[k];

	      //printf("def_ii=%x, allocaed %d, %d\n",def_ii,def_ii->forward_inst, def_inst);
              edge=calloc(1,sizeof(forward_edge));
	      edge->next=(def_ii->edges)[def_inst];
	      edge->ii=ii;
	      edge->inst=k;
              (def_ii->edges)[def_inst]=edge;
          }//for k
       }//for j

/*
#ifdef CONSIDER_POTENTIAL
	
       if (ii->pds) {
	   for (j=0;j<ii->pds_alloced;j++) {
		pd=ii->pds[j];
	        while (pd) {
		   def_ii=pd->pred_ii;
		   def_inst=pd->pred_inst;
	           edge=calloc(1,sizeof(forward_edge));
	           edge->next=(def_ii->edges)[def_inst];
	           edge->ii=ii;
	           edge->inst=j;
                   (def_ii->edges)[def_inst]=edge;
		   pd=pd->next;
		} //while
	   }//for j
	}// if (ii->pds
#endif
*/
    }
    printf("CONSTRUCTED\n");
}     
	    
	      
  

void construct_forward_pdg()
{
    Int i,j,k,def_inst; 
    instr_info * ii, * def_ii;
    val_stream * vs;
    forward_edge * edge;
    potential_dep * pd;

    for (i=0;i<instr_gid;i++) {
        ii=iis[i];
	if (!ii) continue;
        for (j=0;j<ii->use_port;j++) {
          vs=ii->uports[j];
	  if (!vs) continue;
          for (k=0;k<vs->used;k++) {
              def_ii=vs->iis[k];
	      if (!def_ii) continue;
              def_inst=vs->insts[k];
			
	      if (def_inst>=def_ii->forward_inst) def_ii->forward_inst=def_inst+1;
         }
       }
   }
   printf ("Calculated\n");
   for (i=0;i<instr_gid;i++) {
        ii=iis[i];
	if (!ii) continue;
	if (ii->forward_inst < ii->cur_inst) ii->forward_inst=ii->cur_inst;	
        if (ii->forward_inst) {
	   ii->edges=calloc(ii->forward_inst, sizeof(forward_edge*));
           ii->forward_visited=calloc(ii->forward_inst/32+1, sizeof(UInt));
#ifdef DEP_DISTANCE
	   ii->forward_dist=calloc(ii->forward_inst, sizeof(UInt));
#endif     
        } 
		
    }
    printf("ALLOCATED\n"); 
    for (i=0;i<instr_gid;i++) {
        ii=iis[i];
	if (!ii) continue;
        for (j=0;j<ii->use_port;j++) {
          vs=ii->uports[j];
	  if (!vs) continue;
          for (k=0;k<vs->used;k++) {
              def_ii=vs->iis[k];
	      if (!def_ii) continue;
              def_inst=vs->insts[k];

	      //printf("def_ii=%x, allocaed %d, %d\n",def_ii,def_ii->forward_inst, def_inst);
              edge=calloc(1,sizeof(forward_edge));
	      edge->next=(def_ii->edges)[def_inst];
	      edge->ii=ii;
	      edge->inst=k;
              (def_ii->edges)[def_inst]=edge;
          }//for k
       }//for j

/*
#ifdef CONSIDER_POTENTIAL
	
       if (ii->pds) {
	   for (j=0;j<ii->pds_alloced;j++) {
		pd=ii->pds[j];
	        while (pd) {
		   def_ii=pd->pred_ii;
		   def_inst=pd->pred_inst;
	           edge=calloc(1,sizeof(forward_edge));
	           edge->next=(def_ii->edges)[def_inst];
	           edge->ii=ii;
	           edge->inst=j;
                   (def_ii->edges)[def_inst]=edge;
		   pd=pd->next;
		} //while
	   }//for j
	}// if (ii->pds
#endif
*/
    }
    printf("CONSTRUCTED\n");
}     
void delete_slice_list(slice_node * list)
{
    slice_node * old;
    while (list) {
       old=list;
       list=list->next;
       free(old);
    }
}
slice_node * get_all_output(Char* fname)
{
    FILE * fp;
    Char buf[1024];
    fp=fopen(fname,"r");
    Int ii_id,inst;
    slice_node * result=NULL;
    while (fgets(buf,1024,fp)) {
	if (buf[0]=='&' && buf[1]=='*') {
	   int i=3;
	   while (buf[i]!=',') i++;
	   buf[i]='\0';
	   ii_id=atoi(buf+3);
	   inst=atoi(buf+i+1);
           result=insert_node_into_list(iis[ii_id],inst, result);
        }
    }
    return result;
}
#ifdef MAPPING
Int get_mapping(instr_info * ii, UInt val)
{
    val_mapping * mapping;
    val_link * link;
    Int cnt,upper,lower,mid;
    Int * indices;
    UInt t;
    Int index,i,j;
    mapping=ii->mapping;
    if (ii->id==4183) { 
	printf("XXXXX 4183\n");
    }
    if (!mapping) return 1;
    cnt=mapping->size;
    indices=mapping->indices;
    lower=0;
    upper=cnt-1;
    mid=(lower+upper)/2;
    t=indices[mid];
     while (lower!=mid && val!=t) {
	      if (t>val) {
		  upper=mid;
	      } else {
		  lower=mid;
	      }
	      mid=(upper+lower)/2;
              t=indices[mid];
     }
            
     
     if (val==t) {
        index=mid;
	//return mapping->links[mid]->cnt;
     } else if (val==indices[upper])
        index=upper;
        //return mapping->links[upper]->cnt;
     link=mapping->links[index];
     cnt=link->cnt;

     if (ii->id==4183) {
	 printf("XXXXX 4183 val=%d, index=%d, cnt=%d\n",val,index,cnt);
     }

     //heuristics
     /*
     if (cnt>1) {
	for (i=0;i<cnt;i++) {
	   for (j=0;j<link->uport;j++) {
	     if (link->expanded[j][i]>0x10000000) {//it is a address
		return 1;
             }
	   }//for j
        }
     }
     */
     //not found the value
     return cnt;
}
    
void construct_mapping ()
{   
    Int i,j,k,l,unique_cnt,upper,lower,mid,def_inst;
    val_mapping * mapping;
    val_link* link;
    UInt val,t;
    UInt unique_vals[MAX_MAPPING_ENTRY];
    UInt dep_val[MAX_MAPPING_PORT];
    instr_info * ii, * def_ii;
    val_stream * vs;
    //val_link indices[MAX_MAPPING_ENTRY];
    
    //get sorted unique values;
    for (i=0;i<instr_gid;i++) {
	ii=iis[i];
	printf("compute mapping for %d, allocated=%d\n",i,ii->allocated);
        if (!ii || !ii->values || ii->use_port==1 || ii->use_port>MAX_MAPPING_PORT) continue;

        unique_cnt=1;
        unique_vals[0]=ii->values[0];
        for (k=1;k<ii->allocated;k++) {
	   lower=0;
	   upper=unique_cnt-1;
	   val=ii->values[k];
	   mid=(lower+upper)/2;
	   t=unique_vals[mid];
	   while (lower!=mid && val!=t) {
	      if (t>val) {
		  upper=mid;
	      } else {
		  lower=mid;
	      }
	      mid=(upper+lower)/2;
              t=unique_vals[mid];
           }
            
	   if (val==t || val==unique_vals[upper]) continue;
           if (val>unique_vals[upper]) mid=upper+1;
           else if (val>t) mid++;
	   //if (mid!=upper && val>unique_vals[upper])mid++;
	   if (unique_cnt>=MAX_MAPPING_ENTRY) break;
	   for (j=unique_cnt;j>mid;j--) {
		   unique_vals[j]=unique_vals[j-1];
	   }
           unique_cnt++;
	   unique_vals[mid]=val;
          /* 
           for (k=0;k<unique_cnt;k++) {
 	      printf("%x ",unique_vals[k]);
           }
           printf("\nmid=%d\n",mid);
        */

	}

	if (unique_cnt>=MAX_MAPPING_ENTRY) continue;

        //allocate the mapping
        mapping=ii->mapping=calloc(1,sizeof(val_mapping));
        mapping->size=unique_cnt;
	mapping->indices=calloc(unique_cnt, sizeof(UInt));
	memcpy(mapping->indices, unique_vals, unique_cnt*sizeof(UInt));
        mapping->links=calloc(unique_cnt, sizeof(val_link*));
        for (k=0;k<unique_cnt;k++) {
	   link=mapping->links[k]=calloc(1, sizeof(val_link));
	   link->uport=ii->use_port-1;
        }
	   
        for (k=0;k<ii->allocated;k++) {
	   lower=0;
	   upper=unique_cnt-1;
	   val=ii->values[k];
	   mid=(lower+upper)/2;
	   t=unique_vals[mid];
	   while (lower!=mid && val!=t) {
	      if (t>val) {
		  upper=mid;
	      } else {
		  lower=mid;
	      }
	      mid=(upper+lower)/2;
              t=unique_vals[mid];
           }
	   

	   //if (val!=t && val!=unique_vals[upper]) {
	   //     printf("val(%x)!=t(%x) upper=%x\n",val,t, unique_vals[upper]);
                assert(val==t||val==unique_vals[upper]);
           //}
           if (val==unique_vals[upper]) mid=upper;
	   link=mapping->links[mid];

	   //already enough diversity 
	   if (link->cnt==MAX_LINK) continue;

           //get the depdent values
	   memset(dep_val,0,MAX_MAPPING_PORT* sizeof(UInt));
	   for (j=1;j<ii->use_port;j++) {
	      vs=ii->uports[j];
	      if (!vs) continue;
	      if (vs->used<=k) continue;
	      def_ii=vs->iis[k];
	      if (!def_ii||!def_ii->values) continue;
              def_inst=vs->insts[k];
	      if (def_inst>=def_ii->allocated) continue;
	      dep_val[j-1]=def_ii->values[def_inst];
	   }
           //no inputs have been found before

           /*
           if (!link->cnt) {
		memcpy(link->vals,dep_val, link->uport*sizeof(UInt));
		link->cnt=link->alloced=1;
           }
	   */

           //find if it appears before
	   Bool found=False;
	   for (j=0;j<link->cnt;j++) {
	       Bool identical=True;
	       for (l=0;l<link->uport;l++) {
		   if (link->expanded[l][j]!=dep_val[l]) {
		     identical=False;
		     break;
		   }
	       }
	       //same tuple has been found to produce the same result
	       if (identical) {
		  found=True;
		  break;
	       }
	   }

	   if (!found) {
	      if (!link->expanded[0]) {
		 for (l=0;l<link->uport;l++) {
		    link->expanded[l]=calloc(MAX_LINK,sizeof(UInt));
		    //link->expanded[l][0]=link->vals[l];
		 }
              }

              for (l=0;l<link->uport;l++) {
		 link->expanded[l][link->cnt]=dep_val[l];
	      }
	      link->cnt++;
           }//if (!found)
         }//for k
      }//for i
}

void write_mapping(Char * name)
{		
   FILE * fp;
   int i,j,k,l;
   val_mapping * mapping;
   val_link * link;
   instr_info * ii;

   fp=fopen(name,"w");
   fprintf(fp,"%d\n",instr_gid);
   for (i=0;i<instr_gid;i++) {
      ii=iis[i];
      fprintf (fp, "%d %x %s %s %d %d\n", ii->id, ii->orig_addr, ii->fn, ii->file, ii->line, ii->use_port);
      mapping=ii->mapping;
      if (!mapping) {
	 fprintf(fp,"NO MAPPING\n");
         continue;
      }
   
      fprintf(fp,"MAPPING %d\n",mapping->size);
      for (j=0;j<mapping->size;j++) {
	 link=mapping->links[j];
         fprintf(fp,"\t%x: %d %d\n",mapping->indices[j],link->cnt, link->uport);
	 if (link->cnt) {
	     for (k=0;k<link->cnt;k++) {
                 fprintf(fp,"\t\t");  
	         for (l=0;l<link->uport;l++) {
		   fprintf(fp,"%x ", link->expanded[l][k]);
		 }
		 fprintf(fp,"\n");
             }
          }/* 
          else {
	     fprintf(fp,"\t\t");
	     for (l=0;l<link->uport;l++) {
		fprintf(fp,"%x ", link->vals[l]);
             }
	     fprintf(fp,"\n");
         } */
      }//for j
   }//for i
   fclose(fp);
}
	 
void read_mapping(Char * name)
{
    FILE * fd=fopen(name,"r");
    Int i,j,k,l,ii_id,size,cnt,uport,offset;
    val_mapping * mapping;
    val_link * link;
    instr_info * ii;
    UInt index;
    Char buf[1024];
    Int port_idx;
        
    if (fd < 0) {
       printf("FATAL: can't open traced file `%s'", name);
       exit(-1);
    }
    fgets(buf,1024,fd); //instr_gid

    while (fgets(buf,1024,fd)) {
      sscanf(buf,"%d ", &ii_id);
      ii=iis[ii_id];
      fgets(buf,1024,fd);
      if (buf[0]=='N') continue; //no mapping

      //fgets(buf,256,fd);
      sscanf(buf,"MAPPING %d",&size);
      mapping=ii->mapping=calloc(1, sizeof(val_mapping));
      mapping->size=size;
      mapping->indices=calloc(size, sizeof(UInt));
      mapping->links=calloc(size, sizeof(val_link*));

      for (j=0;j<size;j++) {
	 fgets(buf,1024,fd);
         printf("%s",buf);
	 i=0;
	 offset=i;
         while (buf[i]!=':') i++;
	 buf[i]='\0';
	 index=strtoll(buf+offset,NULL,16);
 
         i+=2;
         offset=i;
         while (buf[i]!=' ') i++;
         buf[i]='\0';
         cnt=atoll(buf+offset);
        
         i++;
         offset=i;
         while (buf[i]!=' '&& buf[i]!='\n') i++;
         buf[i]='\0';
         uport=atoll(buf+offset);

         printf("READ: %x %d %d\n",index, cnt, uport);
 
 
         //sscanf(buf+i+1,"%d %d",&cnt,&uport);

	 //sscanf(buf,"\t%x: %d %d", &index, &cnt, &uport);
	 mapping->indices[j]=index;
	 link=mapping->links[j]=calloc(1, sizeof(val_link));
	 link->cnt=cnt;
         link->uport=uport;
    /*     
         if (link->cnt==1) {
	    fgets(buf,256,fd);
            i=0;
            for (l=0;l<uport;l++) {
                offset=i;
                while (buf[i]!=' ' && buf[i]!='\n') i++;
                buf[i]='\0';
                link->vals[l]=strtoll(buf+offset,NULL,16);
		i++;
	    }
         } else {
*/
	    for (l=0;l<uport;l++) {
		link->expanded[l]=calloc(link->cnt, sizeof(UInt));
	    }
	    for (k=0;k<link->cnt;k++) {
	         fgets(buf,1024,fd);
                 i=0;
                 for (l=0;l<uport;l++) {
                    offset=i;
                    while (buf[i]!=' ' && buf[i]!='\n') i++;
                    buf[i]='\0';
                    link->expanded[l][k]=strtoll(buf+offset,NULL,16);
		    i++;
	         }//for l
             } //for k
  //       }//else
      }//for j
   }//while
   fclose(fd);

}	

Int get_ts(instr_info * ii, Int inst)
{
    instr_info * bb;
    bb=ii->bb;
    if (!bb || !bb->tss) return -1;
    return (bb->tss[inst]);
}

//#define LOAD_DEP_UPORT 2
#define MAX_CNT 300
//for relevant slicing
void construct_potential_dependence()
{
    Int i,j,k,u,l,m,def_inst;
    instr_info * ii, * def_ii, *pred_ii, * branch_ii,*temp_ii;
    potential_cd * pcd;
    val_stream * vs;
    potential_dep * pd;
    //potential_cd * pcd;
    Int use_ts, def_ts, tmp_ts,cnt;

    for (i=0;i<instr_gid;i++) {
      ii=iis[i];
      if (!ii->pcd) {//it is not a load with potential cd
	 continue;
      }
      printf("%x(%s,%d) has pcd\n", ii->orig_addr, ii->file, ii->line);
      pcd=ii->pcd;
      //the last port should be the memory dependence
      vs=ii->uports[ii->use_port-1];
      //vs=ii->uports[LOAD_DEP_UPORT];
      if (!vs) continue;
      ii->pds_alloced=vs->used;
      ii->pds=calloc(vs->used, sizeof (potential_dep*));
      printf("%x(%s,%d) has instances\n", ii->orig_addr, ii->file, ii->line);
      //for each instance 
      for (j=0;j<vs->used;j++) {       
         def_ii=vs->iis[j];
         if (!def_ii) continue;
         def_inst=vs->insts[j];
         use_ts=get_ts(ii, j);
         def_ts=get_ts(def_ii, def_inst);
         printf("%x(%s,%d) has use_ts=%d, def_ts=%d\n", ii->orig_addr, ii->file, ii->line,use_ts, def_ts);
         cnt=0;
         //looks at the potential CD
         for (k=0;k<pcd->pcd_cnt;k++) {
	     pred_ii = pcd->p_pred_bbs[k];
	     branch_ii = pcd->p_branches[k];
             //printf("%x(%s,%d) PCD pred %x(%s,%d), branch %x(%s,%d)\n", ii->orig_addr, ii->file, ii->line, pred_ii->orig_addr, pred_ii->file, pred_ii->line, branch_ii->orig_addr, branch_ii->file, branch_ii->line);
             if (!pred_ii || !pred_ii->tss) continue;
             u=pred_ii->ts_alloced;
             l=0;
             m=u/2;
             while (u-l>1) {
                //tmp_ts=get_ts(pred_ii,m);
		pred_ii->tss[m];
		if (tmp_ts > def_ts) {
		    u=m;
		    m=(u+l)/2;    
		} else {
		    l=m;
		    m=(u+l)/2;
                }
             }
             while (tmp_ts<use_ts && m<pred_ii->ts_alloced) {
	       temp_ii=get_instr_info_by_pc(pred_ii->cfs[m]);
               if (temp_ii) {
                     //printf("pred %x(%s,%d), inst=%d, cfs[m]=%x(%s,%d)\n", pred_ii->orig_addr, pred_ii->file, pred_ii->line, m, temp_ii->orig_addr, temp_ii->file, temp_ii->line);
               } else {
                     //printf("pred %x(%s,%d), inst=%d, NULL cfs[m]\n", pred_ii->orig_addr, pred_ii->file, pred_ii->line, m);
               }
		
	       if (temp_ii !=branch_ii ) {
                   //satisfy the condition, add it as a potential dep
                   pd=calloc(1, sizeof (potential_dep));
		   pd->pred_ii=pred_ii->pred_ii;
		   pd->pred_inst=m;
		   pd->next=ii->pds[j];
		   ii->pds[j]=pd;
		   /*
		   if (pd->pred_ii) 
		       printf("PD: (%x,%s,%d),%d --> (%x,%s,%d),%d\n", pd->pred_ii->orig_addr, pd->pred_ii->file, pd->pred_ii->line, m, ii->orig_addr, ii->file, ii->line, j);
		   */
		   cnt++;
	           if (cnt>MAX_CNT) {
			printf("WARNING: the number of potential dep exceeds MAXfor load %x (%s,%d)\n",ii->orig_addr, ii->file, ii->line);
		   }
                }
		m++;
	        //tmp_ts=get_ts(pred_ii,m);
		tmp_ts=pred_ii->tss[m];
	     }
          }//k
       }//j
    }//i
		
}

//the potential edges for switching is orgnaized as follows,
//each pe is a predicate, it has a link list of use (load) iis which potentiallydepend on this predicate

#define PES_HASH_ENTRY 4096
sw_potential_edge_head * pes=NULL;
sw_potential_edge_head * pes_table[PES_HASH_ENTRY];

Int ii_inst_hash(instr_info * ii, Int inst)
{
    UInt hash=ii->orig_addr;
    hash+=inst;
    return (hash%4096);
}

sw_potential_edge_head * get_pe (instr_info * pred_ii, Int pred_inst)
{
    Int idx=ii_inst_hash(pred_ii,pred_inst);
    sw_potential_edge_head * pes_entry=pes_table[idx];
    Bool found=False; 
    while (pes_entry) {
       if (pes_entry->pred_ii==pred_ii && pes_entry->inst==pred_inst) {
           return pes_entry;
       }
       pes_entry=pes_entry->hash_next;
    }
    pes_entry=calloc(1, sizeof (sw_potential_edge_head));
    pes_entry->pred_ii=pred_ii;
    pes_entry->inst=pred_inst;
    pes_entry->ts=get_ts(pred_ii,pred_inst);

    //add it to the global list
    pes_entry->next=pes;
    pes=pes_entry; 
    //add it to the hash table
    pes_entry->hash_next=pes_table[idx];
    pes_table[idx]=pes_entry;
    return pes_entry;
}


#define MAX_TAILS 1024
sw_potential_edge_tail * tails[MAX_TAILS];

static int tail_compare(const void * n1, const void * n2)
{
    sw_potential_edge_tail * t1=*((sw_potential_edge_tail**)n1);
    sw_potential_edge_tail * t2=*((sw_potential_edge_tail**)n2);
    Int ts1=t1->ts;
    Int ts2=t2->ts;
    if (ts1< ts2) return -1;
    else 1;
    //if (sn1->inst<sn2->inst) return -1;
    //else return 1;
}

void write_potential_edge_candidates(Int err_ii_id, Int err_inst, Int num_expand)
{
    FILE * fp;
    Char buf[128];
    sprintf(buf,"pe_%d_%d_%d.candidates",err_ii_id, err_inst, num_expand);
    fp=fopen(buf,"w");
    sw_potential_edge_head * temp_pe=pes;
    sw_potential_edge_tail * tail;
    Int idx,i;
    instr_info * ii;
    while (temp_pe) {
	ii=temp_pe->pred_ii;
	fprintf(fp,"%x %d %d (%s,%s,%d)\n", ii->orig_addr, temp_pe->inst, temp_pe->ts, ii->file, ii->fn, ii->line);
        tail=temp_pe->tails;
        idx=0;
        while (tail&&idx<MAX_TAILS) {
           tails[idx++]=tail;
	   tail=tail->next;
        }
        qsort(tails,idx, sizeof (sw_potential_edge_tail *), tail_compare);
        for (i=0;i<idx;i++) {
	     tail=tails[i];
             ii=tail->use_ii;
	     fprintf(fp,"\t%x %d %d %x (%s,%s,%d)\n", ii->orig_addr, tail->inst, 
tail->ts, tail->value, ii->file, ii->fn, ii->line);
        }
        temp_pe=temp_pe->next;
     }
     fclose(fp);
}

//load in the predicates which have been proved being really  potentially dependent on
//num_expand means how many steps have been proved 
slice_node * read_proved_potential_edges (Int err_ii_id, Int err_inst, Int num_expand)
{
    FILE * fp;
    Char buf[256];
    Addr ii_addr;
    Int inst,i,j;
    instr_info * ii;
    slice_node * result=NULL;
    for (j=0;j<num_expand-1;j++) {
       sprintf(buf,"pe_%d_%d_%d.proved\n",err_ii_id,err_inst, j+1);
       fp=fopen(buf,"r");
       if (fp<0) return result;
       while (fgets(buf,256,fp)) {
           i=0;
	   while (buf[i]!=' ') i++;
	   buf[i]='\0';
           ii_addr=strtoll(buf,NULL,16);
	   inst=atoll(buf+i+1);
           ii=get_instr_info_by_pc(ii_addr);
           result=insert_node_into_list(ii,inst,result);
       }
       fclose(fp);
    }//for (j
    return result;
}	

static Bool no_parent_has_been_visited(instr_info * ii, Int inst)
{
        Int i,tmp_inst;
	val_stream * vs;
        instr_info * tmp_ii;
        for (i=1;i<ii->use_port;i++) {
            vs=ii->uports[i];
	    
            if (vs&&vs->used>inst && vs->iis[inst]) {
	        tmp_ii=vs->iis[inst];
	        tmp_inst=vs->insts[inst];
                if (get_vs_visited_bit(tmp_ii,tmp_inst)) {
		   return False;
		}
	    }
        }
	return True;
}

/* the visited bits must not be cleared before calling this function*/
slice_node * expand_once (slice_node * slice, Int err_ii_id, Int err_inst, Int num_expand,Bool proved, slice_node* (* slice_body) (slice_node *))
{
    slice_node * criteria=NULL, * expanded_slice;
    instr_info * ii=NULL, * pred_ii=NULL;
    potential_dep * pd= NULL;
    slice_node * list=slice;
    Int inst,pred_inst;
    sw_potential_edge_head * pe;
    sw_potential_edge_tail * pe_tail;

    //get the set of statements that are potentially dependent on
    while (list) {
        ii=list->ii;
        inst=list->inst;
        if (!ii->pds || ii->pds_alloced<=inst) {
	   list=list->next;
	   continue;
        }

//#define AGGRESSIVE_CERTIFY
#ifdef AGGRESSIVE_CERTIFY
        if (!no_parent_has_been_visited(ii,inst)) {
	   list=list->next;
	   continue;
        }
#endif
        /*
	if (ii->line==503 && inst==26 || (ii->line==511 && inst==25)) {
		list=list->next;
		continue;
	}
        */
	if (list->step<20) {
	    list=list->next;
	    continue;
        }


        pd=ii->pds[inst];
        while (pd) {
           pred_ii=pd->pred_ii;
	   pred_inst=pd->pred_inst;
           //if (!get_vs_visited_bit(pred_ii,pred_inst) && no_parent_has_been_visited(pred_ii,pred_inst)) {
           if (!get_vs_visited_bit(pred_ii,pred_inst) ) {
	       criteria=insert_node_into_list(pred_ii,pred_inst,criteria);
               set_vs_visited_bit(pred_ii,pred_inst);
		   //if (pd->pred_ii) 
		       printf("PD: (%x,%s,%d),%d  --> (%x,%s,%d),%d s=%d\n", pred_ii->orig_addr, pred_ii->file, pred_ii->line, pred_inst, ii->orig_addr, ii->file, ii->line, inst,list->step);
               
	       if (proved) { //indicate it is for switching
	          pe=get_pe(pred_ii,pred_inst);
	          pe_tail=calloc(1,sizeof (sw_potential_edge_tail));
	          pe_tail->use_ii=ii;
	          pe_tail->inst=inst;
	          pe_tail->ts=get_ts(ii,inst);
	          pe_tail->next=pe->tails;
		  if (ii->values && ii->allocated>inst) {
		     pe_tail->value=ii->values[inst];
                  } else  pe_tail->value=-1;
	          pe->tails=pe_tail;
	       }
           }
	   pd=pd->next;
        }
        list=list->next;
     }
     if (proved) write_potential_edge_candidates(err_ii_id, err_inst, num_expand); 
     //slicing from the potential set
     expanded_slice=(*slice_body)(criteria);
     //expanded_slice=slice_from_traced_cut(criteria);
    
     //combine the new slice and the old one.
     list=expanded_slice;
     while (list) {
         slice=insert_node_into_list(list->ii, list->inst, slice);
         list=list->next;
     }
     
     return slice;
}

//starting from the output, reach the cut whose files are traced
slice_node * relevant_slice(slice_node * work_list)
{
     slice_node * result=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp, * old;
     potential_dep * pd;

     instr_info * tmp_ii;
     Int tmp_inst;
     Int traversed=0;

     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
        traversed++;
        if (traversed % 1000==0) {
	   //if (traversed>10000) exit(-1); 
           //printf("TRAVERSED %d\n",traversed);
        }
#undef DEBUG
//#define DEBUG
#ifdef DEBUG
        printf("process %d(%x,%s,%s,%d) %d v=%x'%c'\n",ii->id,ii->orig_addr, ii->file, ii->fn, ii->line,inst,work_list->value,work_list->value&0x7f);
#endif
	old=work_list;
        work_list=work_list->next;
        if (is_file_traced(ii->file,ii->fn,ii->line)) {
            //temp=create_slice_node(ii,inst);
            old->next=NULL;
            result=insert_node_into_list(ii,inst, result);
            //work_list=work_list->next;
            //continue;
        }
	//work_list=work_list->next;
#ifdef DEBUG
        //printf("\t use_port=%d\n",ii->use_port);
#endif
        for (i=0;i<ii->use_port;i++) {
            vs=ii->uports[i];
            if (vs) {
#ifdef DEBUG
	       //printf("\tvs has %d elements\n",vs->used);
#endif
               if (vs->used>inst && vs->iis[inst]) {

		  tmp_ii=vs->iis[inst];
	          tmp_inst=vs->insts[inst];
                  if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\tinsert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	                 set_vs_visited_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
		  }//if
               }
            }
         }//for

         //potential dep
         if (ii->pds && ii->pds_alloced > inst) {
	     pd=ii->pds[inst];
             while (pd) {
		tmp_ii=pd->pred_ii;
	        tmp_inst=pd->pred_inst;
                if (!get_vs_visited_bit(tmp_ii,tmp_inst)) {
                       //temp=create_slice_node(vs->iis[inst], vs->insts[inst]);	     
#ifdef DEBUG
                         printf("\tinsert to wl %d(%s,%s,%d) %d\n",tmp_ii->id,tmp_ii->file,tmp_ii->fn,tmp_ii->line, tmp_inst);
#endif
    	                 set_vs_visited_bit(tmp_ii,tmp_inst);
                         work_list=insert_node_into_list(tmp_ii, tmp_inst, work_list);
		}//if
	        pd=pd->next;
             }//while
          }//if (potential dep)
               
	 free(old);
#ifdef DEBUG
	//printf("WORKLIST:\n");
        //print_slice_list(work_list);
#endif
     }
     return result;
#undef DEBUG
}



#endif
