#include "utility.h"
#include "lk_helper.h"
Int ALLOC_STEPS[4]={1,10,100,300};
//Int ALLOC_STEPS[4]={1,10,100,10000};
#define STEP_COUNT 4
Int cd_fun_entry_count;
cd_fun_entry ** cd_fun_entries;
cd_fun_entry *dummy_cd_fun_entry;

Bool user_phase=False;

/*X.Z.*/
VgHashTable instr_info_table;
VgHashTable fun_entry_table;

#define MAX_INSTR_GID 1000000
static instr_info * iis[MAX_INSTR_GID];
Int instr_gid=0;

// Top level of CC table.  Auto-zeroed.
static fileCC *file_CC_table[N_FILE_ENTRIES];
static traced_file *traced_file_table[N_FILE_ENTRIES];

//function entries
//VgHashTable fun_entry_table;
//VgHashTable bb_entry_table;

Dep_Buffer_Element * depBuffer[DEP_BUFFER_SIZE] = {0};
UInt nextDepBufferSlot = 0;
void initDepBuffer()
{
   Int i;
   for(i = 0; i < DEP_BUFFER_SIZE; i++)
   {
      Dep_Buffer_Element *element = (Dep_Buffer_Element *)(VG_(malloc)(sizeof(Dep_Buffer_Element)));
      element->isUsed = False;
      depBuffer[i] = element;
   }
}


void get_debug_info(Addr instr_addr, Char file[FILE_LEN],
                           Char fn[FN_LEN], Int* line)
{
   Bool found_file_line = VG_(get_filename_linenum)(instr_addr, file,
                                                    FILE_LEN, line);
   Bool found_fn        = VG_(get_fnname)(instr_addr, fn, FN_LEN);

   if (!found_file_line) {
      VG_(strcpy)(file, "???");
      *line = 0;
   }
   if (!found_fn) {
      VG_(strcpy)(fn,  "???");
   }
}

static UInt hash(Char *s, UInt table_size)
{
   const int hash_constant = 256;
   int hash_value = 0;
   for ( ; *s; s++)
      hash_value = (hash_constant * hash_value + *s) % table_size;
   return hash_value;
}

static __inline__
fileCC* new_fileCC(Char filename[], fileCC* next)
{
   // Using calloc() zeroes the fns[] array
   fileCC* cc = VG_(calloc)(1, sizeof(fileCC));
   cc->file   = VG_(strdup)(filename);
   cc->next   = next;
   return cc;
}
static __inline__
traced_file* new_traced_file(Char filename[], traced_file * next)
{
   // Using calloc() zeroes the fns[] array
   traced_file* tf = VG_(calloc)(1, sizeof(traced_file));
   tf->file   = VG_(strdup)(filename);
   tf->next   = next;
   return tf;
}

static __inline__
fnCC* new_fnCC(Char fn[], fnCC* next)
{
   // Using calloc() zeroes the lines[] array
   fnCC* cc = VG_(calloc)(1, sizeof(fnCC));
   cc->fn   = VG_(strdup)(fn);
   cc->next = next;
   return cc;
}

static Bool is_file_traced(Char * file)
{
      traced_file * curr_traced_file;
      Int file_hash;
      file_hash = hash(file, N_FILE_ENTRIES);
      curr_traced_file   = traced_file_table[file_hash];
      while (NULL != curr_traced_file && !VG_STREQ(file, curr_traced_file->file)) {
         curr_traced_file = curr_traced_file->next;
      }
      return (curr_traced_file!=NULL);
}

Bool is_instr_traced(Addr orig_addr)
{
   Char    file[FILE_LEN], fn[FN_LEN];
   Int     line;
   get_debug_info(orig_addr,file,fn,&line);
   return is_file_traced(file);
}

static Int cd_fun_entry_cmp(void * cd1, void * cd2)
{
    cd_fun_entry * t1=*(cd_fun_entry**) cd1;
    cd_fun_entry * t2=*(cd_fun_entry**) cd2;
    return (t1->entry_addr < t2->entry_addr? -1:1);
}

void read_cd_fun_entries (Char * fname)
{
   Int fd,i,idx,eof;
   Addr addr;
   cd_fun_entry * entry;
   Char buf[1024];
   fd = VG_(open)( fname, VKI_O_RDONLY, 0 );
   //VG_(printf) ("READ CONTROL DEP ENTRY FOR\n");
   if (fd < 0)
   {
      // Allow program to run without control dependence info. 
      //VG_(message)(Vg_UserMsg, "WARNING: can't read control dep file `%s'", fname);
      //VG_(exit)(1);
      return;
   }
   VG_(get_line) (fd, buf,1024);
   cd_fun_entry_count=VG_(atoll) (buf);
   cd_fun_entries=VG_(calloc) (cd_fun_entry_count, sizeof (cd_fun_entry*));
   dummy_cd_fun_entry=VG_(calloc) (1, sizeof(cd_fun_entry));
   

   idx=0;
   while (True) {
      eof = VG_(get_line) ( fd, buf, 1024 );
      if (eof) break;
      i=0;
      while (buf[i]!=' ') i++;
      cd_fun_entries[idx]=entry=VG_(calloc) (1, sizeof(cd_fun_entry));
      entry->entry_addr=VG_(atoll16)(buf);
      VG_(strcpy) (entry->fn, buf+i+1);
      idx++;
   }
   VG_(close) (fd);
   VG_(ssort) ((void*) cd_fun_entries, cd_fun_entry_count, sizeof(cd_fun_entry*), cd_fun_entry_cmp);
 }


fun_entry * get_fun_entry_new(Addr addr, fileCC * filecc, fnCC * fncc)
{
  Int ub, lb, mark;
  cd_fun_entry * entry;
  fun_entry * fe;

  lb=0; 
  ub=cd_fun_entry_count;
  mark=(ub+lb)/2;

  if(cd_fun_entries == NULL)
  {
     return get_fun_entry_dummy(filecc, fncc);
  }
  
  entry=cd_fun_entries[mark];
  while (mark!=lb) {
    if (entry->entry_addr==addr) {
	 break;
    } else if (entry->entry_addr> addr) {
         ub=mark;
    } else {
         lb=mark;
    }
    mark=(ub+lb)/2;
    
    entry=cd_fun_entries[mark];
  }
  if (VG_(strcmp)(entry->fn, fncc->fn)) {// CD is not available for this function
      //VG_(printf) ("CD not available, search for %x, %s, found %x, %s\n", addr, fncc->fn, entry->entry_addr, entry->fn);
      return get_fun_entry_dummy(filecc, fncc);
      //return dummy_cd_fun_entry;
  } else {
      if (!entry->fe) {
	fe=entry->fe=VG_(calloc) (1, sizeof(fun_entry));
        fe->orig_addr=addr;
        fe->file=filecc;
        fe->fn=fncc;

        //fe->cd_available=0;
        fe->cd_available=read_control_dep (entry->entry_addr, fncc->fn, fe);
      }
      //VG_(printf) ("FOUND cd %x, %s at %x\n", addr, fncc->fn,entry->entry_addr);
      return entry->fe;
 }
}
    
#define MAX_BB_NUM 8192
static instr_info * bbs [MAX_BB_NUM];

static Int instr_info_cmp (void * be1, void * be2)
{
  instr_info * t1=*(instr_info **) be1;
  instr_info * t2=*(instr_info **) be2;
  return (t1->orig_addr<t2->orig_addr ? -1 : 1);
}

#define MAX_CD 20
#define LV_BUF 2

#define TO_LOWER(buf) \
     for (i=0;i<VG_(strlen)(buf);i++) { \
	if (buf[i]>='A' && buf[i]<='Z') { \
	   buf[i]-= 'A'-'a'; \
        } \
     }

/*the instruction info for the first instruction of BBs are created*/
static Bool read_control_dep(Addr fun_addr, Char * fn, fun_entry * fe)
{
   Char buf[1024];
   Char * cd;
   Int fd,i;
   UInt pc;
   Int cd_cnt=0;
   Int pred_cnt=0;
   UInt cds[MAX_CD];
   //bb_entry * be, * tmp_be;
   instr_info * be, * tmp_be;
   Bool eof;
   Int bb_cnt=0;
   Bool new_be=False;

   //method strpbrk doesn't have the move %esp, %ebp
   if (!VG_(strcmp)(fn,"strpbrk") || !VG_(strcmp)(fn,"strlen") ||
      !VG_(strcmp)(fn,"memcpy") || !VG_(strcmp)(fn,"_IO_doallocbuf")
      || !VG_(strcmp)(fn,"_IO_vfprintf")
      || !VG_(strcmp)(fn,"strrchr")
      || !VG_(strcmp)(fn,"strcmp")
      || !VG_(strcmp)(fn,"_IO_new_fdopen")
      || !VG_(strcmp)(fn,"memset")
      || !VG_(strcmp)(fn,"__ieee754_log")
      || !VG_(strcmp)(fn,"memmove")
      || !VG_(strcmp)(fn,"memcmp")
      || !VG_(strcmp)(fn,"strchr")
      || !VG_(strcmp)(fn,"strcat")
      || !VG_(strcmp)(fn,"__mpn_mul_1")
/* //for flex
      || !VG_(strcmp)(fn,"_IO_new_file_write")
      || !VG_(strcmp)(fn,"_IO_new_file_underflow")
      || !VG_(strcmp)(fn,"new_do_write")
      || !VG_(strcmp)(fn,"_IO_file_doallocate")
*/
     ) return False;
   VG_(sprintf) (buf,"./cd/%x.cd",fun_addr);
   TO_LOWER(buf);
   fd = VG_(open)( buf, VKI_O_RDONLY, 0 );
   //VG_(printf) ("READ CONTROL DEP FOR (%x)%s\n",fun_addr, fn);
   if (fd < 0) {
      VG_(message)(Vg_UserMsg, "WARNING: can't read control dep file `%s'",
                   buf);
      //VG_(exit)(1);
      return False;
   }
   
   while (True) {
      eof = VG_(get_line) ( fd, buf, 1024 );
      if (eof) break;
          i=0;
          while (True) {
                  if (buf[i]==':') {
                          buf[i-1]='\0';
                          break;
                  }
                  if (buf[i]=='\0') {
                          goto syntax_error;
                  }
                  i++;
          }
          pc=VG_(atoll16) (buf);
	  //get_bb_info(pc, &cur_fileCC, & cur_fnCC, & line);
          be=get_instr_info(pc, fe, INVALID_LINE, INVALID_UPORT, &new_be,False);
          if (new_be) {
              bbs[bb_cnt++]=be;
          }

          //get the CDs
          i+=2;
          cd=buf+i;
          cd_cnt=0;
          //VG_(printf) ("buf[%d]=%s\n",i,buf+i);
          while (buf[i]!='\0') {
                 if (buf[i]=='-') {
                          buf[i]='\0';
                          cds[cd_cnt++]=VG_(atoll16)(cd);
                          //cd=buf+i+1;
                          i++;
                          while (buf[i]!=' ') i++;
                          cd=buf+i+1;

                 }
                 i++;
          }
          if (cd!=buf+i) {
                cds[cd_cnt++]=VG_(atoll16)(cd);
          }
          be->cd_cnt=cd_cnt;
          if (cd_cnt) {//(be->cds!=NULL) means the cds has been established
              be->cds=VG_(calloc) (cd_cnt, sizeof(instr_info*));
              for (i=0;i<cd_cnt;i++) {
                 tmp_be=be->cds[i]=get_instr_info(cds[i],fe,INVALID_LINE, INVALID_UPORT,&new_be, False);
                 if (new_be) {
                    bbs[bb_cnt++]=tmp_be;
		    //tmp_be->lv_idx=pred_cnt;
		    //pred_cnt++;
                 }
		 if (tmp_be->lv_idx==LV_NOT_ASSIGNED) {
		    tmp_be->lv_idx=pred_cnt;
		    pred_cnt++;
                 }
              }
              //pred_cnt++;
          }
   }
   
   //have not been assigned, first time read
   fe->alloc_lv=pred_cnt+LV_BUF;
   //fe->alloc_lv=pred_cnt;
   //VG_(printf)("fe->alloc_lv=%d\n",fe->alloc_lv);
   VG_(ssort) ((void*) bbs, bb_cnt, sizeof(instr_info*), instr_info_cmp);
   for (i=0;i<bb_cnt-1;i++) {
       //VG_(printf) ("Ordered:%x\n",bbs[i]->orig_addr);
       bbs[i]->next_bb=bbs[i+1];
       bbs[i]->bb=bbs[i];
       //bbs[i]->lv_idx=i;
   }
   if (bb_cnt>0) {
      bbs[bb_cnt-1]->next_bb=NULL;
      bbs[bb_cnt-1]->bb=bbs[bb_cnt-1];
   }
   
   fe->first_bb=bbs[0];
   VG_(close)(fd);
   return (pred_cnt>0); 

syntax_error:
   VG_(message)(Vg_UserMsg,
                   "FATAL: in control dep file: `%s': syntax error on: %s",
                   fn, buf );
   VG_(close)(fd);
   VG_(message)(Vg_UserMsg, "exiting now.");
   VG_(exit)(1);

}

instr_info * search_for_bb(Addr orig_addr, fun_entry * fe)
{
   instr_info * bb,* last_bb;
   last_bb=NULL;
   bb=fe->first_bb;
   //VG_(printf)("first_bb=%x\n",bb);
   //VG_(printf)("first_bb=%x\n",bb->orig_addr);
   while (bb && orig_addr >= bb->orig_addr) {
      //VG_(printf) ("try to search %x, try %x\n", orig_addr, bb->orig_addr);
      last_bb=bb;
      bb=bb->next_bb;
   }
   //VG_(printf) ("Search %x, get bb %x\n", orig_addr, last_bb->orig_addr);
   return last_bb;
}
      
   

fun_entry * get_fun_entry_dummy(fileCC * filecc, fnCC * fncc)
{
      fun_entry * fe;
      VgHashNode ** dummy;
      fe=(fun_entry*) VG_(HT_get_node) (fun_entry_table, fncc, &dummy);
      if (fe==NULL) {
         //fe->first_ii=ii;
         fe=(fun_entry*) VG_(calloc) (1, sizeof(fun_entry));
         //fe->orig_addr=func_addr;
         fe->file=filecc;
         fe->fn=fncc;
         VG_(HT_add_node)( fun_entry_table, (VgHashNode*)fe );
	 //fe->cd_available= read_control_dep(func_addr, fncc->fn, fe);
      }
      return fe;
}

//static __inline__ 
void get_bb_info(Addr bb_addr, fileCC ** filecc, fnCC ** fncc, Int *line)
{
   // Using calloc() zeroes instr_size and data_size
   //instr_info* ii;
   //VgHashNode** dummy;
   Char    file[FILE_LEN], fn[FN_LEN];
   //Int     line;
   UInt    file_hash, fn_hash;
   //Int i;
   fileCC * curr_fileCC;
   fnCC * curr_fnCC;
  

   get_debug_info(bb_addr,file,fn,line);

   //VG_(printf) ("%x %s %s %d\n", bb_addr,file,fn, *line);

      file_hash = hash(file, N_FILE_ENTRIES);
      curr_fileCC   = file_CC_table[file_hash];
      while (NULL != curr_fileCC && !VG_STREQ(file, curr_fileCC->file)) {
         curr_fileCC = curr_fileCC->next;
      }
      if (NULL == curr_fileCC) {
         file_CC_table[file_hash] = curr_fileCC =
                  new_fileCC(file, file_CC_table[file_hash]);
      }

      fn_hash = hash(fn, N_FN_ENTRIES);
      curr_fnCC   = curr_fileCC->fns[fn_hash];
      while (NULL != curr_fnCC && !VG_STREQ(fn, curr_fnCC->fn)) {
          curr_fnCC = curr_fnCC->next;
      }
      if (curr_fnCC==NULL) {
         curr_fileCC->fns[fn_hash] = curr_fnCC =
                  new_fnCC(fn, curr_fileCC->fns[fn_hash]);
      } 
      *filecc=curr_fileCC;
      *fncc=curr_fnCC;
      //VG_(printf) ("good\n");
}

//static __inline__ 
instr_info* get_instr_info(Addr instr_addr, fun_entry * fe,Int line, Int uport, Bool * new_instr, Bool look_for_bb)
{
   // Using calloc() zeroes instr_size and data_size
   instr_info* ii;
   VgHashNode** dummy;
  

   ii = (instr_info*)VG_(HT_get_node)(instr_info_table, instr_addr, &dummy);
   if (ii) {
	 //VG_(printf)("OLD INSTR %x",instr_addr);
	 if (ii->line==INVALID_LINE) {
	     //ii->use_port=uport;
	     ii->line=line;
         }
	 * new_instr=False;
         return ii;
   }
   *new_instr=True;


	 //VG_(printf)("NEW INSTR %x",instr_addr);
      ii = VG_(calloc)(1, sizeof(instr_info));
      ii->orig_addr = instr_addr;
      if (look_for_bb) { 
	  //VG_(printf) ("SEARCH FOR BB\n");
	  ii->bb=search_for_bb(instr_addr, fe);
      }
      //ii->instr_size=instr_size;
      //ii->data_size=data_size;

      sk_assert(instr_gid<MAX_INSTR_GID);
      iis[instr_gid]=ii;
      ii->id=instr_gid++;
     //VGP_PUSHCC(VgpGetLineCC);

      ii->file=fe->file->file;
      ii->fn=fe->fn->fn;
      ii->fe=fe;
      ii->line=line;

      
      ii->lv_idx=LV_NOT_ASSIGNED; 
     
      //ii->bb_addr=bb_addr;
      ii->use_port=uport;
      //ii->parent     = parent; 
      VG_(HT_add_node)( instr_info_table, (VgHashNode*)ii );

#ifdef PRINT_DETAIL
      VG_(printf) ("XXXX Instr %d: <%s, %s, %d>\n",ii->id, ii->file, ii->fn, ii->line);
#endif
      //return ii->id;
   return ii;
}

/*
//static __inline__ 
instr_info* get_instr_info(Addr instr_addr,Addr bb_addr, Int len)
{
   // Using calloc() zeroes instr_size and data_size
   instr_info* ii;
   VgHashNode** dummy;
   Char    file[FILE_LEN], fn[FN_LEN];
   Int     line;
   UInt    file_hash, fn_hash;
   Int i;
   Bool is_traced;
   fileCC * curr_fileCC;
   fnCC * curr_fnCC;
  

   ii = (instr_info*)VG_(HT_get_node)(instr_info_table, instr_addr, &dummy);
   if (ii) return ii;

   get_debug_info(instr_addr,file,fn,&line);
   //is_traced=is_file_traced(file);

   //if (!is_traced) {
   //	return NULL;
   //}
   user_phase=True;

      file_hash = hash(file, N_FILE_ENTRIES);
      curr_fileCC   = file_CC_table[file_hash];
      while (NULL != curr_fileCC && !VG_STREQ(file, curr_fileCC->file)) {
         curr_fileCC = curr_fileCC->next;
      }
      if (NULL == curr_fileCC) {
         file_CC_table[file_hash] = curr_fileCC =
                  new_fileCC(file, file_CC_table[file_hash]);
         distinct_files++;
      }

      fn_hash = hash(fn, N_FN_ENTRIES);
      curr_fnCC   = curr_fileCC->fns[fn_hash];
      while (NULL != curr_fnCC && !VG_STREQ(fn, curr_fnCC->fn)) {
          curr_fnCC = curr_fnCC->next;
      }
      if (curr_fnCC==NULL) {
         curr_fileCC->fns[fn_hash] = curr_fnCC =
                  new_fnCC(fn, curr_fileCC->fns[fn_hash]);
         distinct_fns++;
      } 


      ii = VG_(calloc)(1, sizeof(instr_info));
      ii->orig_addr = instr_addr;
      //ii->instr_size=instr_size;
      //ii->data_size=data_size;

      sk_assert(instr_gid<MAX_INSTR_GID);
      iis[instr_gid]=ii;
      ii->id=instr_gid++;
     //VGP_PUSHCC(VgpGetLineCC);

      ii->file=curr_fileCC->file;
      ii->fn=curr_fnCC->fn;
      ii->line=line;
      ii->bb_addr=bb_addr;
      ii->len=len;
      //ii->parent     = parent; 
      VG_(HT_add_node)( instr_info_table, (VgHashNode*)ii );

#ifdef PRINT_DETAIL
      VG_(printf) ("XXXX Instr %d: <%s, %s, %d>\n",ii->id, ii->file, ii->fn, ii->line);
#endif
      //return ii->id;
   return ii;
}
*/
void load_traced_file ( Char* filename )
{
#  define N_BUF 200
   Int   fd;
   Bool  eof;
   Char  buf[N_BUF+1];
   UInt file_hash;
   traced_file * curr_traced_file;

   fd = VG_(open)( filename, VKI_O_RDONLY, 0 );
   if (fd < 0) {
      VG_(message)(Vg_UserMsg, "FATAL: can't open traced file `%s'",
                   filename );
      VG_(exit)(1);
   }

   while (True) {
      /* Assign and initialise the two suppression halves (core and tool) */
      eof = VG_(get_line) ( fd, buf, N_BUF );
      if (eof) break;

      file_hash = hash(buf, N_FILE_ENTRIES);
      curr_traced_file   = traced_file_table[file_hash];
      while (NULL != curr_traced_file && !VG_STREQ(buf, curr_traced_file->file)) {
         curr_traced_file = curr_traced_file->next;
      }
      if (NULL == curr_traced_file) {
         traced_file_table[file_hash] = curr_traced_file =
             new_traced_file(buf, traced_file_table[file_hash]);
         //distinct_files++;
      }
   }
}

void print_cb(UCodeBlock * cb)
{
        Int i;
        for (i = 0; i < VG_(get_num_instrs)(cb); i++) {
           UInstr * instr=VG_(get_instr)(cb,i);
           if (instr==NULL) {
               VG_(printf)("%d is NULL\n",i);
               continue;
           }
           if (instr->opcode != NOP)
               VG_(pp_UInstr) ( i, instr );
        }
        VG_(printf)("\n");
}

io_log * XZ_logs[OPEN_FILE_MAX];
Int log_count=0;

size_t read_from_log(Int fid, Char * bf, size_t count)
{
    Int i;
    io_log * cur_log;
    Int offset;

    //no interaction
    if (fid==0) VG_(exit)(0);

    for (i=0;i<log_count;i++) {
	if (XZ_logs[i]->fid==fid) {
	   cur_log=XZ_logs[i];
	   break;
	}
    }

    if (i>=log_count) {
	cur_log=XZ_logs[log_count++]=VG_(calloc) (1, sizeof(io_log));
	cur_log->fid=fid;
	//VG_(printf)("RESTART\n");
        VG_(lseek)(fid,0, VKI_SEEK_SET);
	cur_log->count=VG_(read) (fid, cur_log->buf, IO_BUF_MAX);
	//if (cur_log->count<count) count=cur_log->count;
    }
        
    if (cur_log->offset==cur_log->count) {
	//no interaction
	//if (fid==0) VG_(exit)(0);
        cur_log->file_pos+=cur_log->count;
	offset=VG_(lseek) (fid, cur_log->file_pos, VKI_SEEK_SET);
	if (offset!=cur_log->file_pos) return 0;
 	cur_log->count=VG_(read) (fid,cur_log->buf, IO_BUF_MAX);
	cur_log->offset=0;
    }

    count= (cur_log->offset+count<=cur_log->count)?count:cur_log->count-cur_log->offset;
    VG_(memcpy) (bf, cur_log->buf+cur_log->offset, count);
    cur_log->offset+=count;

    //VG_(printf)("fid=%d, return %d\n",fid,count);
    return count;
}

val_stream * create_val_stream()
{
    val_stream * vs;
    vs=VG_(calloc)(1, sizeof(val_stream));
    vs->allocated=ALLOC_STEPS[0];
    vs->iis=VG_(calloc)(vs->allocated, sizeof(UInt));
    vs->insts=VG_(calloc)(vs->allocated, sizeof(UInt));
    return vs;
}

static val_stream * create_val_stream_with_size(Int size)
{
    val_stream * vs;
    vs=VG_(calloc)(1, sizeof(val_stream));
    vs->allocated=size;
    vs->iis=VG_(calloc)(vs->allocated, sizeof(UInt));
    vs->insts=VG_(calloc)(vs->allocated, sizeof(UInt));
    return vs;
}

#ifdef TRACK_INPUT
//the relevance with input
void set_input_relevance(instr_info * ii,UInt value)
{
    Int new_size,i;
    Int * temp;
    Int use_inst=ii->cur_inst;
    new_size=-1;
    //sk_assert(use_inst-vs->used<10);
    while (use_inst+1>=ii->input_related_alloc) {
	 if (ii->input_related_alloc>= ALLOC_STEPS[STEP_COUNT-1]) {
	     new_size=ii->input_related_alloc+ALLOC_STEPS[STEP_COUNT-1];
         } else {
             for (i=0;i<STEP_COUNT-1;i++) if (ii->input_related_alloc<=ALLOC_STEPS[i]) break;
	     new_size=ALLOC_STEPS[i+1];
         }
	 ii->input_related_alloc=new_size;
    }
    //VG_(printf)("Try to resize\n");
    if (new_size!=-1) {
	 if (ii->input_related) { 
             ii->input_related=VG_(realloc) (ii->input_related, new_size*sizeof(UInt));
             VG_(memset) (ii->input_related+use_inst, 0xfffe, (new_size-use_inst)*sizeof(UInt));
	 } else 
             ii->input_related=VG_(calloc) (new_size, sizeof(UInt));
	 //VG_(printf) ("use_inst=%d, vs->allocated=%d, new_size=%d\n", use_inst, vs->allocated, new_size);
    }
    ii->input_related[use_inst]=value;
    //vs->used=use_inst+1;
}
#endif

#ifdef TIMESTAMPED
//expand value stream to accomodate use_inst
void insert_timestamp(instr_info * ii, UInt ts)
{
    Int new_size,i;
    Int * temp;
    
    if(VG_(doOnlyBufferOutput)() == False)
    {
       Int use_inst=ii->cur_inst-1;
       //Int use_inst=ii->cur_inst;
       new_size=-1;
       //VG_(printf)("INSERT TS bb=%x (add=%x)\n",ii,ii->orig_addr);
       //sk_assert(use_inst-vs->used<10);
       while (use_inst>=ii->ts_alloced) {
   	 if (ii->ts_alloced>= ALLOC_STEPS[STEP_COUNT-1]) {
   	     new_size=ii->ts_alloced+ALLOC_STEPS[STEP_COUNT-1];
            } else {
                for (i=0;i<STEP_COUNT-1;i++) if (ii->ts_alloced<=ALLOC_STEPS[i]) break;
   	     new_size=ALLOC_STEPS[i+1];
            }
   	 ii->ts_alloced=new_size;
       }
       //VG_(printf)("use_inst=%d, new_size=%d\n",use_inst, new_size);
       //VG_(printf)("Try to resize\n");
       if (new_size!=-1) {
   	 if (ii->tss) { 
                ii->tss=VG_(realloc) (ii->tss, new_size*sizeof(UInt));
                ii->cfs=VG_(realloc) (ii->cfs, new_size*sizeof(Addr));
                //VG_(memset) (ii->values+use_inst, 0xfffe, (new_size-use_inst)*sizeof(Int));
   	 } else { 
                ii->tss=VG_(calloc) (new_size, sizeof(UInt));
                ii->cfs=VG_(calloc) (new_size, sizeof(Addr));
   	 }
   	 //VG_(printf) ("use_inst=%d, vs->allocated=%d, new_size=%d\n", use_inst, vs->allocated, new_size);
       }
       //VG_(printf)("\t\tINSERT VALUE %x @ <%x,%d>\n",value, ii->orig_addr, use_inst);
       ii->tss[use_inst]=ts;
       //VG_(printf)("INSERT TS done\n");
       //vs->used=use_inst+1;
    }
}

void insert_control_flow(instr_info * bb, Addr branch_pc)
{
    if(VG_(doOnlyBufferOutput)() == False)
    {
       //VG_(printf)("INSERT CF bb=%x (addr=%x)\n",bb,bb->orig_addr);
       sk_assert (bb->cfs);

       //This happens when there are more than one bb in one trace
       //if (!bb->cfs || bb->ts_alloced<=bb->cur_inst-1) return;

       //this happens when the basic block has only one instruction
           


       //use_inst has been incremented
       bb->cfs[bb->cur_inst-1]=branch_pc;
       //VG_(printf)("INSERT CF done\n");
    }
}     
#endif

//expand value stream to accomodate use_inst
void insert_value(instr_info * ii, UInt value)
{
    // Xiangyu code below
    if(VG_(doOnlyBufferOutput)() == False)
    {
       Int new_size,i;
       Int * temp;
       Int use_inst=ii->cur_inst-1;
       //Int use_inst=ii->cur_inst;
       new_size=-1;
       //sk_assert(use_inst-vs->used<10);
       while (use_inst+1>=ii->allocated) {
   	 if (ii->allocated>= ALLOC_STEPS[STEP_COUNT-1]) {
   	     new_size=ii->allocated+ALLOC_STEPS[STEP_COUNT-1];
            } else {
                for (i=0;i<STEP_COUNT-1;i++) if (ii->allocated<=ALLOC_STEPS[i]) break;
   	     new_size=ALLOC_STEPS[i+1];
            }
   	 ii->allocated=new_size;
       }
       //VG_(printf)("Try to resize\n");
       if (new_size!=-1) {
   	 if (ii->values) { 
                ii->values=VG_(realloc) (ii->values, new_size*sizeof(UInt));
                //VG_(memset) (ii->values+use_inst, 0xfffe, (new_size-use_inst)*sizeof(Int));
   	 } else 
                ii->values=VG_(calloc) (new_size, sizeof(UInt));
   	 //VG_(printf) ("use_inst=%d, vs->allocated=%d, new_size=%d\n", use_inst, vs->allocated, new_size);
       }
       //VG_(printf)("\t\tINSERT VALUE %x @ <%x,%d>\n",value, ii->orig_addr, use_inst);
       ii->values[use_inst]=value;
       //vs->used=use_inst+1;
    }
}
//expand value stream to accomodate use_inst
static void expand_val_stream(val_stream ** p_vs,Int use_inst)
{
    Int new_size,i;
    val_stream * vs= *p_vs;
    Int * temp;
    //used=vs->used;
    //VG_(printf) ("insert dependence use_inst=%d\n",use_inst);
    //return;
    //if (1) return;
    if (vs==NULL) { 
	vs=*p_vs=create_val_stream();
    }
    new_size=-1;
    //sk_assert(use_inst-vs->used<10);
    while (use_inst+1>=vs->allocated) {
	 if (vs->allocated>= ALLOC_STEPS[STEP_COUNT-1]) {
	     new_size=vs->allocated+ALLOC_STEPS[STEP_COUNT-1];
         } else {
             for (i=0;i<STEP_COUNT-1;i++) if (vs->allocated==ALLOC_STEPS[i]) break;
	     new_size=ALLOC_STEPS[i+1];
         }
	 vs->allocated=new_size;
    }
    //VG_(printf)("Try to resize %d\n",new_size);
    if (new_size!=-1) {
         if (vs->iis) {
 	     vs->iis=VG_(realloc) (vs->iis, new_size*sizeof(Int));
             VG_(memset) (vs->iis+vs->used, 0, (new_size-vs->used)*sizeof(Int));
         } else
             vs->iis=VG_(calloc) (new_size, sizeof(Int));

	 if (vs->insts) { 
             vs->insts=VG_(realloc) (vs->insts, new_size*sizeof(Int));
             VG_(memset) (vs->insts+vs->used, 0, (new_size-vs->used)*sizeof(Int));
	 } else 
             vs->insts=VG_(calloc) (new_size, sizeof(Int));
/*
	 if (vs->values) { 
             vs->values=VG_(realloc) (vs->values, new_size*sizeof(Int));
             VG_(memset) (vs->values+vs->used, 0xfffe, (new_size-vs->used)*sizeof(Int));
	 } else 
             vs->values=VG_(calloc) (new_size, sizeof(Int));
*/
	 //VG_(printf) ("use_inst=%d, vs->allocated=%d, new_size=%d\n", use_inst, vs->allocated, new_size);
    }
    vs->used=use_inst+1;
}


void insert_dependence(val_stream ** p_vs, instr_info * ii,Int use_inst, instr_info * def_ii, Int def_inst)
{
    if(VG_(doOnlyBufferOutput)() == True)
    {
       depBuffer[nextDepBufferSlot]->isUsed = True;
       //depBuffer[nextDepBufferSlot]->useInstrID = ii->id;
       depBuffer[nextDepBufferSlot]->useInstrID = ii->orig_addr;
       depBuffer[nextDepBufferSlot]->useInst = use_inst;
       //depBuffer[nextDepBufferSlot]->defInstrID = def_ii->id;
       depBuffer[nextDepBufferSlot]->defInstrID = def_ii->orig_addr;
       depBuffer[nextDepBufferSlot]->defInst = def_inst;

       /*
       VG_(printf)("%d:%d depends upon %d:%d\n", depBuffer[nextDepBufferSlot]->useInstrID,
                                                 depBuffer[nextDepBufferSlot]->useInst = use_inst,
                                                 depBuffer[nextDepBufferSlot]->defInstrID = def_ii->id,
                                                 depBuffer[nextDepBufferSlot]->defInst = def_inst);
       */

       nextDepBufferSlot = (nextDepBufferSlot+1) % DEP_BUFFER_SIZE;
    }
    else
    {
       // Xiangyu code below
       expand_val_stream(p_vs, use_inst);
       val_stream * vs = *p_vs;
       vs->iis[use_inst] = (UInt)def_ii;
       vs->insts[use_inst] = def_inst;
    }
}

void write_buffer_trace(Char * file)
{
   Int fd, i;
   Char buf[1024];
   
   fd=VG_(open)(file, VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY, VKI_S_IRUSR|VKI_S_IWUSR);
   if (fd < 0)
   {
      VG_(message)(Vg_UserMsg, "FATAL: can't open output file `%s'", file);
      VG_(exit)(1);
   }
   
   for(i = 0; i < DEP_BUFFER_SIZE; i++)
   {
      if(depBuffer[i]->isUsed == True)
      {
         VG_(sprintf) (buf, "0x%x#%d --> 0x%x#%d\n", depBuffer[i]->useInstrID, depBuffer[i]->useInst, depBuffer[i]->defInstrID, depBuffer[i]->defInst);
         VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
      }
   }
   
   VG_(close)(fd);
}

void write_instr_dic ( Char * file)
{
   Int fd,i;
   instr_info * ii;
   Char buf[1024];
   Int bb_id, pred_id;

   fd=VG_(open)(file, VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                                       VKI_S_IRUSR|VKI_S_IWUSR);
   
   if (fd < 0) {
      VG_(message)(Vg_UserMsg, "FATAL: can't open traced file `%s'",
                   file );
      VG_(exit)(1);
   }
   VG_(sprintf) (buf, "%d\n",instr_gid);
   VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
   
   VG_(sprintf) (buf, "ID  PC  BB  FN  FILE  LINE USE_PORT JMP_PC IS_PREDICATE PRED_II\n");
   VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
   for (i=0;i<instr_gid;i++) {
      ii=iis[i];
      bb_id=-1;
      if (ii->bb) bb_id=ii->bb->id;
      pred_id=-1;
      if (ii->pred_ii) pred_id=ii->pred_ii->id;
      VG_(sprintf) (buf, "%d %x %d %s %s %d %d %x %d %d\n", ii->id, ii->orig_addr, bb_id, ii->fn, ii->file, ii->line, ii->use_port,ii->jmp_pc,ii->is_predicate, pred_id);
//      else VG_(sprintf) (buf, "%d %x %d %s %s %d %d %x %d %x\n", ii->id, ii->orig_addr, -1, ii->fn, ii->file, ii->line, ii->use_port,ii->jmp_pc, ii->is_predicate, ii->pred_ii);
      VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
   }
   VG_(close)(fd);
}

static instr_info * create_instr_info (Addr pc, Int bb_id, Char * fn, Char * file, Int line, Int id, Int uport)
{
     instr_info * ii;
     ii=VG_(calloc) (1, sizeof(instr_info));

     ii->orig_addr=pc;
     ii->bb_id=bb_id;

     ii->fn=VG_(calloc) (VG_(strlen)(fn)+1, sizeof (Char));
     VG_(strcpy) (ii->fn, fn);

     ii->file=VG_(calloc) (VG_(strlen)(file)+1, sizeof (Char));
     VG_(strcpy) (ii->file, file);

     ii->line=line;
     ii->id=id;
     ii->use_port=uport;
     return ii;
}

void read_instr_dic (Char * dic)
{
    Int fd, i, offset,eof,line,id,uport,bb_id;
    Char buf[256];
    Char fn[128];
    Char file[128];
    Addr pc;    
    instr_info * ii;
    fd = VG_(open)( dic, VKI_O_RDONLY, 0 );
    if (fd < 0) {
       VG_(message)(Vg_UserMsg, "FATAL: can't open traced file `%s'",
                   dic);
       VG_(exit)(1);
    }

    VG_(get_line) (fd,buf,256);
    instr_gid=VG_(atoll) (buf);
    //instr_gid=0;
 
    VG_(get_line) (fd,buf,256);

    while (True) {
      /* Assign and initialise the two suppression halves (core and tool) */
      eof = VG_(get_line) ( fd, buf, 256);
      //VG_(printf)("%s\n",buf);
      if (eof) break;

      i=0;
      offset=i; 
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      id=VG_(atoll)(buf+offset);
 
      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      pc=VG_(atoll16)(buf+offset);

  
      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      bb_id=VG_(atoll)(buf+offset);

      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      VG_(strcpy)(fn, buf+offset);

      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      VG_(strcpy)(file, buf+offset);

      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      line=VG_(atoll)(buf+offset);
     
      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      uport=VG_(atoll)(buf+offset);

      iis[id]=create_instr_info(pc, bb_id, fn, file, line, id, uport);
   }
   VG_(close)(fd);

   for (i=0;i<instr_gid;i++) {
       ii=iis[i];
       if (ii->bb_id!=-1) ii->bb=iis[ii->bb_id];
   }
}

static void write_val_stream (Int fd, val_stream * vs)
{
   Char buf[256];
   Int i;
   instr_info * ii;
   if (!vs || !VG_(is_addressable) (vs,sizeof(val_stream))
   ){
       VG_(sprintf) (buf, "SIZE 0\n");
       VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
       return;
   }
    
   VG_(sprintf) (buf, "SIZE %d\n",vs->used);
   VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
   //DEBUG
   //VG_(printf)("%s",buf);
   //VG_(sprintf) (buf, "\tII INST\n");
   //VG_(write) (fd, (void*)buf, VG_(strlen)(buf));

   for (i=0;i<vs->used;i++) {
        //DEBUG
        //VG_(printf)("i=%d, iis=%x\n",i,vs->iis);
        if (!VG_(is_addressable) ((vs->iis+i),4) ||
          !VG_(is_addressable) ((vs->insts+i),4)) {
             VG_(printf)("WARNING: CORRUPTED\n");
             VG_(sprintf) (buf, "\t%d:-1 -1\n",i);
        } else {
           ii=(instr_info*) vs->iis[i];
           //DEBUG
          // VG_(printf)("iis[i]=%x\n",ii);
	   if (ii==NULL || !VG_(is_addressable)(ii,sizeof(instr_info))) {
	      // VG_(printf) ("!ii\n"); 
               VG_(sprintf) (buf, "\t%d:-1 -1\n",i);
   	   //if (!ii) VG_(sprintf) (buf, "-1 -1\n");
           } else {
	       //VG_(printf) ("ii!=NULL\n");
               VG_(sprintf) (buf, "\t%d:%d %d\n",i,ii->id,vs->insts[i]);
           }
        }
	//DEBUG
        //VG_(printf) ("DONE:%s",buf);

        VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
   }
   
}
  /* 
static val_stream * read_val_stream (fd)
{
   val_stream * vs;
   Int size,i,j,offset, ii_id;
   Char buf[256];

   VG_(get_line) (fd,buf,256);//"SIZE %d"
   size=VG_(atoll) (buf+5);
   if (!size) return NULL;
   vs=create_val_stream_with_size(size);
    
   VG_(get_line) (fd,buf,256);//"II INST VALUE"

   for (i=0;i<size;i++) {
      VG_(get_line) (fd,buf,256);//"%d %d %d"

      j=0;
      offset=j; 
      while (buf[j]!=' ' && buf[j]!='\n') j++;
      buf[j]='\0';
      ii_id=VG_(atoll) (buf+offset);
      if (ii_id==-1) vs->iis[i]=NULL;
      else vs->iis[i]=(UInt)iis[ii_id];
      vs->values[i]=VG_(atoll16) (buf+j+1);
   }
   return vs;
}
    */
   

void write_dependence ( Char * file)
{
   Int fd,i,j;
   instr_info * ii;
   val_stream * vs;
   Char buf[1024];

   fd=VG_(open)(file, VKI_O_CREAT|VKI_O_TRUNC|VKI_O_WRONLY,
                                       VKI_S_IRUSR|VKI_S_IWUSR);
   
   if (fd < 0) {
      VG_(message)(Vg_UserMsg, "FATAL: can't open traced file `%s'",
                   file );
      VG_(exit)(1);
   }
   VG_(sprintf) (buf, "%d\n",instr_gid);
   VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
   
   //VG_(sprintf) (buf, "ID  USE_PORT PC FILE FN LINE\n");
   //VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
   for (i=0;i<instr_gid;i++) {
      ii=iis[i];
      //could be corrupted by memory error
      if (!VG_(is_addressable)) (ii,sizeof(instr_info));
      VG_(sprintf) (buf, "%d %d %x %s %s %d\n", ii->id, ii->use_port,ii->orig_addr, ii->file, ii->fn, ii->line);
      VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
      //DEBUG
      //VG_(printf) ("%s",buf);
      for (j=0;j<ii->use_port;j++) {
         //DEBUG
         //VG_(printf)("\t port %d\n", j);
	 vs=ii->uports[j];
         write_val_stream(fd,vs);
      }
//#ifdef VALUE_WANTED
      if (!ii->values && !ii->input_related) {
         VG_(sprintf) (buf, "NO VALUES\n");
         VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
      } else if (ii->values && !ii->input_related) {
	 VG_(sprintf) (buf, "VALUES %d\n",ii->cur_inst);
         VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
         for (j=0;j<ii->cur_inst;j++) {
            VG_(sprintf) (buf, "\t%d:%x\n",j,ii->values[j]);
            VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
         }
      } else if (ii->input_related && !ii->values) {
         VG_(sprintf) (buf, "INPUTS %d\n",ii->cur_inst);
         VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
         for (j=0;j<ii->cur_inst;j++) {
            VG_(sprintf) (buf, "\t%d:%d\n",j,ii->input_related[j]);
            VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
         }   
      } else {
	 if (ii->input_related) {
             VG_(sprintf) (buf, "BOTH %d INPUT and VALUES\n",ii->cur_inst);
             VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
             for (j=0;j<ii->cur_inst;j++) {
                VG_(sprintf) (buf, "\t%d:%d %x\n",j,ii->input_related[j],ii->values[j]);
                VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
             }
	 } 
      }
      
      /*
      if (!ii->tss) {
         VG_(sprintf) (buf, "NO TSS\n");
         VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
      } else {
	 VG_(sprintf) (buf, "TSS %d\n",ii->cur_inst);
         VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
         for (j=0;j<ii->cur_inst;j++) {
            VG_(sprintf) (buf, "\t%d:%x %x\n",j,ii->tss[j],ii->cfs[j]);
            VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
         }
      }
      */
	 
//#else
//      VG_(sprintf) (buf, "NO VALUES %d\n");
//      VG_(write) (fd, (void*)buf, VG_(strlen)(buf));
//#endif
         
      //DEBUG
      //VG_(printf) ("DONE\n"); 
   }
   VG_(close)(fd);
}
/*
//require read in the instruction dictionary first
void read_dependence (Char * file)
{
    Int fd, i, offset,eof,id,uport,j;
    Char buf[256];
    instr_info * ii;
    fd = VG_(open)( file, VKI_O_RDONLY, 0 );
    if (fd < 0) {
       VG_(message)(Vg_UserMsg, "FATAL: can't open traced file `%s'",
                   file);
       VG_(exit)(1);
    }
    VG_(get_line) (fd,buf,256); //instr_gid
    //instr_gid=VG_(atoll) (buf);
 
    VG_(get_line) (fd,buf,256); //"ID USE_PORT"

    while (True) {
      eof = VG_(get_line) ( fd, buf, 256);
      //VG_(printf)("%s\n",buf);
      if (eof) break;

      i=0;
      offset=i; 
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      id=VG_(atoll)(buf+offset);
 
      i++;
      offset=i;
      while (buf[i]!=' ' && buf[i]!='\n') i++;
      buf[i]='\0';
      uport=VG_(atoll)(buf+offset);
      
      ii=iis[id];
      for (j=0;j<uport;j++) {
          ii->uports[j]=read_val_stream(fd);
      }
   }//while
}
*/
    
/*     
static slice_node *  insert_node_into_list (slice_node * node, slice_node * work_list)
{
    slice_node * head, * prev;
    head=work_list;
    //rank by (ii->id, inst)
    while (work_list && work_list->ii->id > node->ii->id ) { 
	   prev=work_list;
	   work_list=work_list->next;
    }
    
    while (work_list && work_list->ii->id==node->ii->id && work_list->inst> node->inst) {
	   prev=work_list;
	   work_list=work_list->next;
    }
    //node has bee in the list
    if (work_list && node->ii->id==work_list->ii->id && node->inst==work_list->inst) return work_list;
	
    node->next=work_list;
    if (prev) prev->next=node;
    if (!head) head=node;
    return head;
}

//starting from the output, reach the cut whose files are traced
slice_node * slice_to_traced_cut(slice_node * work_list)
{
     slice_node * result=NULL;
     instr_info * ii=NULL;
     val_stream * vs;
     Int inst,i;
     slice_node * temp;
     
     while (work_list) {
        ii=work_list->ii;
        inst=work_list->inst;
        if (is_file_traced(ii->file)) {
	    temp=create_slice_node(ii,inst);
	    result=insert_node_into_list(temp, result);
	    continue;
        } else {
           for (i=0;i<ii->use_port;i++) {
 	       vs=ii->uport[i];
	       if (vs) {
                   if (vs->used>inst && vs->iis[inst]) {
		       temp=create_slice_node(vs->iis[inst], vs->insts[inst]);
		       work_list=insert_node_into_list(temp,work_list);
                   }
               }
            }//for
        }
	work_list=work_list->next;
     }
     return result;
} 		 	   

void print_slice_list(slice_node * head)
{
     slice_node * list=head;
     while (list) {
	VG_(printf) ("< (%x, %s, %s, %d) , %d >\n", list->ii->orig_addr, list->ii->file, list->ii->fn, list->ii->line, list->inst);
        list=list->next;
     }
}
*/
/*---------------------shadow memory----------------------*/
#define PAGE_SIZE 65536
#define PAGE_NUM 262144

#define IS_DISTINGUISHED_SM(smap) \
   ((smap) == &distinguished_secondary_map)

#define ENSURE_MAPPABLE(map, addr,caller)                              \
   do {                                                           \
      if (IS_DISTINGUISHED_SM(map[(addr) >> 16])) {       \
         map[(addr) >> 16] = alloc_secondary_map(caller); \
         /* VG_(printf)("new 2map because of %p\n", addr); */     \
      }                                                           \
   } while(0)

#define ENSURE_MAPPABLE_BYTE_GRANUITY(map,addr,caller)         \
   do {                                                           \
      if (IS_DISTINGUISHED_SM(map[(addr)&0x03][(addr) >> 16])) {    \
          map[(addr)&0x03][(addr) >> 16] = alloc_secondary_map(caller); \
         /* VG_(printf)("new 2map because of %p\n", addr); */     \
      }                                                           \
   } while(0)



typedef struct {
   UChar byte[PAGE_SIZE];
} SecMap;

static SecMap distinguished_secondary_map;

static SecMap * ii_primary_map[4][PAGE_NUM];
static SecMap * inst_primary_map[4][PAGE_NUM];

#ifdef RELEVANT_SLICING
static SecMap * ts_primary_map[4][PAGE_NUM];
#endif



void init_shadow_memory(void)
{
    Int i,j;
 
    for (i = 0; i< PAGE_SIZE; i++) 
       distinguished_secondary_map.byte[i] = VGM_BYTE_INVALID;
   
    for (j=0;j<4; j++)
      for (i = 0; i < PAGE_NUM; i++) {
        ii_primary_map[j][i] = &distinguished_secondary_map;
        inst_primary_map[j][i] = &distinguished_secondary_map;
#ifdef RELEVANT_SLICING
        ts_primary_map[j][i] = &distinguished_secondary_map;
#endif
      }
}

static SecMap* alloc_secondary_map (__attribute__ ((unused))
                                     Char* caller  )
{
   SecMap* map;
   UInt  i;

   /* Mark all bytes as invalid access and invalid value. */
   map = (SecMap *)VG_(shadow_alloc)(sizeof(SecMap));

   for (i = 0; i < 65536; i++)
      map->byte[i] = VGM_BYTE_INVALID; /* Invalid Value */

   /* VG_(printf)("ALLOC_2MAP(%s)\n", caller ); */
   return map;
}


#ifdef RELEVANT_SLICING
UInt __inline__ get_ts_vbytes4_ALIGNED ( Addr a )
{
   SecMap* sm; 
   sm= ts_primary_map[a&0x3][a>>16];

   UInt    sm_off = a & 0xFFFF;
   return ((UInt*)(sm->byte))[sm_off >> 2];
}


void __inline__ set_ts_vbytes4_ALIGNED ( Addr a, UInt bytes )
{
   SecMap* sm;
   UInt    sm_off;

   ENSURE_MAPPABLE_BYTE_GRANUITY(ts_primary_map,a, "set_ts_vbytes4_ALIGNED");
   sm     = ts_primary_map[a&0x03][a >> 16];

   sm_off = a & 0xFFFF;
   ((UInt*)(sm->byte))[sm_off >> 2] = bytes;
}
#endif

UInt __inline__ get_ii_vbytes4_ALIGNED ( Addr a )
{
   SecMap* sm;
   sm= ii_primary_map[a&0x3][a>>16];

   UInt    sm_off = a & 0xFFFF;
   return ((UInt*)(sm->byte))[sm_off >> 2];
}


void __inline__ set_ii_vbytes4_ALIGNED ( Addr a, UInt bytes )
{
   SecMap* sm;
   UInt    sm_off;
   ENSURE_MAPPABLE_BYTE_GRANUITY(ii_primary_map, a, "set_ii_vbytes4_ALIGNED");
   sm     = ii_primary_map[a&0x03][a >> 16];

   sm_off = a & 0xFFFF;
   ((UInt*)(sm->byte))[sm_off >> 2] = bytes;
}

UInt __inline__ get_inst_vbytes4_ALIGNED ( Addr a )
{
   SecMap* sm;
   sm= inst_primary_map[a&0x3][a>>16];

   UInt    sm_off = a & 0xFFFF;
   return ((UInt*)(sm->byte))[sm_off >> 2];
}


void __inline__ set_inst_vbytes4_ALIGNED ( Addr a, UInt bytes )
{
   SecMap* sm;
   UInt    sm_off;
   ENSURE_MAPPABLE_BYTE_GRANUITY(inst_primary_map,a, "set_inst_vbytes4_ALIGNED");
   sm     = inst_primary_map[a&0x03][a >> 16];

   sm_off = a & 0xFFFF;
   ((UInt*)(sm->byte))[sm_off >> 2] = bytes;
}




